# Copyright 2015 - StackStorm, Inc.
# Copyright 2016 - Brocade Communications Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral.db.v2 import api as db_api
from mistral.services import workbooks as wb_service
from mistral.tests.unit.engine import base
from mistral.workflow import states
from mistral_lib import actions as ml_actions


class SubworkflowPauseResumeTest(base.EngineTestCase):

    def test_pause_resume_cascade_down_to_subworkflow(self):
        workbook = """
        version: '2.0'
        name: wb
        workflows:
          wf1:
            tasks:
              task1:
                workflow: wf2
                on-success:
                  - task3
              task2:
                workflow: wf3
                on-success:
                  - task3
              task3:
                join: all
                action: std.noop
          wf2:
            tasks:
              task1:
                action: std.async_noop
                on-success:
                  - task2
              task2:
                action: std.noop
          wf3:
            tasks:
              task1:
                action: std.async_noop
                on-success:
                  - task2
              task2:
                action: std.noop
        """

        wb_service.create_workbook_v2(workbook)

        # Start workflow execution.
        wf_1_ex = self.engine.start_workflow('wb.wf1')

        self.await_workflow_state(wf_1_ex.id, states.RUNNING)

        with db_api.transaction():
            wf_execs = db_api.get_workflow_executions()

            # Get objects for the parent workflow execution.
            wf_1_ex = self._assert_single_item(wf_execs, name='wb.wf1')

            wf_1_task_execs = wf_1_ex.task_executions

            wf_1_task_1_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task1'
            )

            wf_1_task_1_action_exs = wf_1_task_1_ex.executions

            wf_1_task_2_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task2'
            )

            wf_1_task_2_action_exs = wf_1_task_2_ex.executions

            # Get objects for the subworkflow executions.
            wf_2_ex = self._assert_single_item(wf_execs, name='wb.wf2')

            wf_2_task_execs = wf_2_ex.task_executions

            wf_2_task_1_ex = self._assert_single_item(
                wf_2_ex.task_executions,
                name='task1'
            )

            wf_2_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_task_1_ex.id
            )

            wf_3_ex = self._assert_single_item(wf_execs, name='wb.wf3')

            wf_3_task_execs = wf_3_ex.task_executions

            wf_3_task_1_ex = self._assert_single_item(
                wf_3_ex.task_executions,
                name='task1'
            )

            wf_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_3_task_1_ex.id
            )

        self.assertEqual(states.RUNNING, wf_1_ex.state)
        self.assertEqual(2, len(wf_1_task_execs))
        self.assertEqual(states.RUNNING, wf_1_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_1_task_2_ex.state)
        self.assertEqual(1, len(wf_1_task_1_action_exs))
        self.assertEqual(states.RUNNING, wf_1_task_1_action_exs[0].state)
        self.assertEqual(wf_1_task_1_action_exs[0].id, wf_2_ex.id)
        self.assertEqual(1, len(wf_1_task_2_action_exs))
        self.assertEqual(states.RUNNING, wf_1_task_2_action_exs[0].state)
        self.assertEqual(wf_1_task_2_action_exs[0].id, wf_3_ex.id)
        self.assertEqual(states.RUNNING, wf_2_ex.state)
        self.assertEqual(1, len(wf_2_task_execs))
        self.assertEqual(states.RUNNING, wf_2_task_1_ex.state)
        self.assertEqual(1, len(wf_2_task_1_action_exs))
        self.assertEqual(states.RUNNING, wf_2_task_1_action_exs[0].state)
        self.assertEqual(states.RUNNING, wf_3_ex.state)
        self.assertEqual(1, len(wf_3_task_execs))
        self.assertEqual(states.RUNNING, wf_3_task_1_ex.state)
        self.assertEqual(1, len(wf_3_task_1_action_exs))
        self.assertEqual(states.RUNNING, wf_3_task_1_action_exs[0].state)

        # Pause the main workflow.
        self.engine.pause_workflow(wf_1_ex.id)

        self.await_workflow_paused(wf_1_ex.id)
        self.await_workflow_paused(wf_2_ex.id)
        self.await_workflow_paused(wf_3_ex.id)

        with db_api.transaction():
            wf_execs = db_api.get_workflow_executions()

            # Get objects for the parent workflow execution.
            wf_1_ex = self._assert_single_item(wf_execs, name='wb.wf1')

            wf_1_task_execs = wf_1_ex.task_executions

            wf_1_task_1_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task1'
            )

            wf_1_task_1_action_exs = wf_1_task_1_ex.executions

            wf_1_task_2_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task2'
            )

            wf_1_task_2_action_exs = wf_1_task_2_ex.executions

            # Get objects for the subworkflow executions.
            wf_2_ex = self._assert_single_item(wf_execs, name='wb.wf2')

            wf_2_task_execs = wf_2_ex.task_executions

            wf_2_task_1_ex = self._assert_single_item(
                wf_2_ex.task_executions,
                name='task1'
            )

            wf_2_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_task_1_ex.id
            )

            wf_3_ex = self._assert_single_item(wf_execs, name='wb.wf3')

            wf_3_task_execs = wf_3_ex.task_executions

            wf_3_task_1_ex = self._assert_single_item(
                wf_3_ex.task_executions,
                name='task1'
            )

            wf_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_3_task_1_ex.id
            )

        self.assertEqual(states.PAUSED, wf_2_ex.state)
        self.assertEqual(states.RUNNING, wf_2_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_2_task_1_action_exs[0].state)
        self.assertEqual(states.PAUSED, wf_3_ex.state)
        self.assertEqual(states.RUNNING, wf_3_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_3_task_1_action_exs[0].state)
        self.assertEqual(states.PAUSED, wf_1_task_1_action_exs[0].state)
        self.assertEqual(states.PAUSED, wf_1_task_1_ex.state)
        self.assertEqual(states.PAUSED, wf_1_task_2_action_exs[0].state)
        self.assertEqual(states.PAUSED, wf_1_task_2_ex.state)
        self.assertEqual(states.PAUSED, wf_1_ex.state)

        # Resume the main workflow.
        self.engine.resume_workflow(wf_1_ex.id)

        self.await_workflow_running(wf_1_ex.id)
        self.await_workflow_running(wf_2_ex.id)
        self.await_workflow_running(wf_3_ex.id)

        with db_api.transaction():
            wf_execs = db_api.get_workflow_executions()

            # Get objects for the parent workflow execution.
            wf_1_ex = self._assert_single_item(wf_execs, name='wb.wf1')

            wf_1_task_execs = wf_1_ex.task_executions

            wf_1_task_1_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task1'
            )

            wf_1_task_1_action_exs = wf_1_task_1_ex.executions

            wf_1_task_2_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task2'
            )

            wf_1_task_2_action_exs = wf_1_task_2_ex.executions

            # Get objects for the subworkflow executions.
            wf_2_ex = self._assert_single_item(wf_execs, name='wb.wf2')

            wf_2_task_execs = wf_2_ex.task_executions

            wf_2_task_1_ex = self._assert_single_item(
                wf_2_ex.task_executions,
                name='task1'
            )

            wf_2_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_task_1_ex.id
            )

            wf_3_ex = self._assert_single_item(wf_execs, name='wb.wf3')

            wf_3_task_execs = wf_3_ex.task_executions

            wf_3_task_1_ex = self._assert_single_item(
                wf_3_ex.task_executions,
                name='task1'
            )

            wf_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_3_task_1_ex.id
            )

        self.assertEqual(states.RUNNING, wf_2_ex.state)
        self.assertEqual(states.RUNNING, wf_2_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_2_task_1_action_exs[0].state)
        self.assertEqual(states.RUNNING, wf_3_ex.state)
        self.assertEqual(states.RUNNING, wf_3_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_3_task_1_action_exs[0].state)
        self.assertEqual(states.RUNNING, wf_1_task_1_action_exs[0].state)
        self.assertEqual(states.RUNNING, wf_1_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_1_task_2_action_exs[0].state)
        self.assertEqual(states.RUNNING, wf_1_task_2_ex.state)
        self.assertEqual(states.RUNNING, wf_1_ex.state)

        # Complete action executions of the subworkflows.
        self.engine.on_action_complete(
            wf_2_task_1_action_exs[0].id,
            ml_actions.Result(data={'result': 'foobar'})
        )

        self.engine.on_action_complete(
            wf_3_task_1_action_exs[0].id,
            ml_actions.Result(data={'result': 'foobar'})
        )

        self.await_workflow_success(wf_2_ex.id)
        self.await_workflow_success(wf_3_ex.id)
        self.await_workflow_success(wf_1_ex.id)

        with db_api.transaction():
            wf_execs = db_api.get_workflow_executions()

            # Get objects for the parent workflow execution.
            wf_1_ex = self._assert_single_item(wf_execs, name='wb.wf1')

            wf_1_task_execs = wf_1_ex.task_executions

            wf_1_task_1_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task1'
            )

            wf_1_task_1_action_exs = wf_1_task_1_ex.executions

            wf_1_task_2_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task2'
            )

            wf_1_task_2_action_exs = wf_1_task_2_ex.executions

            wf_1_task_3_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task3'
            )

            # Get objects for the subworkflow executions.
            wf_2_ex = self._assert_single_item(wf_execs, name='wb.wf2')

            wf_2_task_execs = wf_2_ex.task_executions

            wf_2_task_1_ex = self._assert_single_item(
                wf_2_ex.task_executions,
                name='task1'
            )

            wf_2_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_task_1_ex.id
            )

            wf_2_task_2_ex = self._assert_single_item(
                wf_2_ex.task_executions,
                name='task2'
            )

            wf_3_ex = self._assert_single_item(wf_execs, name='wb.wf3')

            wf_3_task_execs = wf_3_ex.task_executions

            wf_3_task_1_ex = self._assert_single_item(
                wf_3_ex.task_executions,
                name='task1'
            )

            wf_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_3_task_1_ex.id
            )

            wf_3_task_2_ex = self._assert_single_item(
                wf_3_ex.task_executions,
                name='task2'
            )

        self.assertEqual(states.SUCCESS, wf_1_ex.state)
        self.assertEqual(3, len(wf_1_task_execs))
        self.assertEqual(states.SUCCESS, wf_1_task_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_1_task_2_ex.state)
        self.assertEqual(states.SUCCESS, wf_1_task_3_ex.state)
        self.assertEqual(states.SUCCESS, wf_1_task_1_action_exs[0].state)
        self.assertEqual(states.SUCCESS, wf_1_task_2_action_exs[0].state)
        self.assertEqual(states.SUCCESS, wf_2_ex.state)
        self.assertEqual(2, len(wf_2_task_execs))
        self.assertEqual(states.SUCCESS, wf_2_task_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_2_task_2_ex.state)
        self.assertEqual(states.SUCCESS, wf_3_ex.state)
        self.assertEqual(2, len(wf_3_task_execs))
        self.assertEqual(states.SUCCESS, wf_3_task_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_3_task_2_ex.state)

    def test_pause_resume_cascade_up_from_subworkflow(self):
        workbook = """
        version: '2.0'
        name: wb
        workflows:
          wf1:
            tasks:
              task1:
                workflow: wf2
                on-success:
                  - task3
              task2:
                workflow: wf3
                on-success:
                  - task3
              task3:
                join: all
                action: std.noop
          wf2:
            tasks:
              task1:
                action: std.async_noop
                on-success:
                  - task2
              task2:
                action: std.noop
          wf3:
            tasks:
              task1:
                action: std.async_noop
                on-success:
                  - task2
              task2:
                action: std.noop
        """

        wb_service.create_workbook_v2(workbook)

        # Start workflow execution.
        wf_1_ex = self.engine.start_workflow('wb.wf1')

        self.await_workflow_state(wf_1_ex.id, states.RUNNING)

        with db_api.transaction():
            wf_execs = db_api.get_workflow_executions()

            # Get objects for the parent workflow execution.
            wf_1_ex = self._assert_single_item(wf_execs, name='wb.wf1')

            wf_1_task_execs = wf_1_ex.task_executions

            wf_1_task_1_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task1'
            )

            wf_1_task_1_action_exs = wf_1_task_1_ex.executions

            wf_1_task_2_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task2'
            )

            wf_1_task_2_action_exs = wf_1_task_2_ex.executions

            # Get objects for the subworkflow executions.
            wf_2_ex = self._assert_single_item(wf_execs, name='wb.wf2')

            wf_2_task_execs = wf_2_ex.task_executions

            wf_2_task_1_ex = self._assert_single_item(
                wf_2_ex.task_executions,
                name='task1'
            )

            wf_2_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_task_1_ex.id
            )

            wf_3_ex = self._assert_single_item(wf_execs, name='wb.wf3')

            wf_3_task_execs = wf_3_ex.task_executions

            wf_3_task_1_ex = self._assert_single_item(
                wf_3_ex.task_executions,
                name='task1'
            )

            wf_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_3_task_1_ex.id
            )

        self.assertEqual(states.RUNNING, wf_1_ex.state)
        self.assertEqual(2, len(wf_1_task_execs))
        self.assertEqual(states.RUNNING, wf_1_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_1_task_2_ex.state)
        self.assertEqual(1, len(wf_1_task_1_action_exs))
        self.assertEqual(states.RUNNING, wf_1_task_1_action_exs[0].state)
        self.assertEqual(wf_1_task_1_action_exs[0].id, wf_2_ex.id)
        self.assertEqual(1, len(wf_1_task_2_action_exs))
        self.assertEqual(states.RUNNING, wf_1_task_2_action_exs[0].state)
        self.assertEqual(wf_1_task_2_action_exs[0].id, wf_3_ex.id)
        self.assertEqual(states.RUNNING, wf_2_ex.state)
        self.assertEqual(1, len(wf_2_task_execs))
        self.assertEqual(states.RUNNING, wf_2_task_1_ex.state)
        self.assertEqual(1, len(wf_2_task_1_action_exs))
        self.assertEqual(states.RUNNING, wf_2_task_1_action_exs[0].state)
        self.assertEqual(states.RUNNING, wf_3_ex.state)
        self.assertEqual(1, len(wf_3_task_execs))
        self.assertEqual(states.RUNNING, wf_3_task_1_ex.state)
        self.assertEqual(1, len(wf_3_task_1_action_exs))
        self.assertEqual(states.RUNNING, wf_3_task_1_action_exs[0].state)

        # Pause the subworkflow.
        self.engine.pause_workflow(wf_2_ex.id)

        self.await_workflow_paused(wf_2_ex.id)

        with db_api.transaction():
            wf_execs = db_api.get_workflow_executions()

            # Get objects for the parent workflow execution.
            wf_1_ex = self._assert_single_item(wf_execs, name='wb.wf1')

            wf_1_task_execs = wf_1_ex.task_executions

            wf_1_task_1_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task1'
            )

            wf_1_task_1_action_exs = wf_1_task_1_ex.executions

            wf_1_task_2_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task2'
            )

            wf_1_task_2_action_exs = wf_1_task_2_ex.executions

            # Get objects for the subworkflow executions.
            wf_2_ex = self._assert_single_item(wf_execs, name='wb.wf2')

            wf_2_task_execs = wf_2_ex.task_executions

            wf_2_task_1_ex = self._assert_single_item(
                wf_2_ex.task_executions,
                name='task1'
            )

            wf_2_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_task_1_ex.id
            )

            wf_3_ex = self._assert_single_item(wf_execs, name='wb.wf3')

            wf_3_task_execs = wf_3_ex.task_executions

            wf_3_task_1_ex = self._assert_single_item(
                wf_3_ex.task_executions,
                name='task1'
            )

            wf_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_3_task_1_ex.id
            )

        self.assertEqual(states.PAUSED, wf_2_ex.state)
        self.assertEqual(states.RUNNING, wf_2_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_2_task_1_action_exs[0].state)
        self.assertEqual(states.PAUSED, wf_3_ex.state)
        self.assertEqual(states.RUNNING, wf_3_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_3_task_1_action_exs[0].state)
        self.assertEqual(states.PAUSED, wf_1_task_1_action_exs[0].state)
        self.assertEqual(states.PAUSED, wf_1_task_1_ex.state)
        self.assertEqual(states.PAUSED, wf_1_task_2_action_exs[0].state)
        self.assertEqual(states.PAUSED, wf_1_task_2_ex.state)
        self.assertEqual(states.PAUSED, wf_1_ex.state)

        # Resume the 1st subworkflow.
        self.engine.resume_workflow(wf_2_ex.id)

        self.await_workflow_running(wf_2_ex.id)

        with db_api.transaction():
            wf_execs = db_api.get_workflow_executions()

            # Get objects for the parent workflow execution.
            wf_1_ex = self._assert_single_item(wf_execs, name='wb.wf1')

            wf_1_task_execs = wf_1_ex.task_executions

            wf_1_task_1_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task1'
            )

            wf_1_task_1_action_exs = wf_1_task_1_ex.executions

            wf_1_task_2_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task2'
            )

            wf_1_task_2_action_exs = wf_1_task_2_ex.executions

            # Get objects for the subworkflow executions.
            wf_2_ex = self._assert_single_item(wf_execs, name='wb.wf2')

            wf_2_task_execs = wf_2_ex.task_executions

            wf_2_task_1_ex = self._assert_single_item(
                wf_2_ex.task_executions,
                name='task1'
            )

            wf_2_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_task_1_ex.id
            )

            wf_3_ex = self._assert_single_item(wf_execs, name='wb.wf3')

            wf_3_task_execs = wf_3_ex.task_executions

            wf_3_task_1_ex = self._assert_single_item(
                wf_3_ex.task_executions,
                name='task1'
            )

            wf_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_3_task_1_ex.id
            )

        self.assertEqual(states.RUNNING, wf_2_ex.state)
        self.assertEqual(states.RUNNING, wf_2_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_2_task_1_action_exs[0].state)
        self.assertEqual(states.PAUSED, wf_3_ex.state)
        self.assertEqual(states.RUNNING, wf_3_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_3_task_1_action_exs[0].state)
        self.assertEqual(states.RUNNING, wf_1_task_1_action_exs[0].state)
        self.assertEqual(states.RUNNING, wf_1_task_1_ex.state)
        self.assertEqual(states.PAUSED, wf_1_task_2_action_exs[0].state)
        self.assertEqual(states.PAUSED, wf_1_task_2_ex.state)
        self.assertEqual(states.PAUSED, wf_1_ex.state)

        # Complete action execution of 1st subworkflow.
        self.engine.on_action_complete(
            wf_2_task_1_action_exs[0].id,
            ml_actions.Result(data={'result': 'foobar'})
        )

        self.await_workflow_success(wf_2_ex.id)
        self.await_task_success(wf_1_task_1_ex.id)

        with db_api.transaction():
            wf_execs = db_api.get_workflow_executions()

            # Get objects for the parent workflow execution.
            wf_1_ex = self._assert_single_item(wf_execs, name='wb.wf1')

            wf_1_task_execs = wf_1_ex.task_executions

            wf_1_task_1_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task1'
            )

            wf_1_task_1_action_exs = wf_1_task_1_ex.executions

            wf_1_task_2_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task2'
            )

            wf_1_task_2_action_exs = wf_1_task_2_ex.executions

            # Get objects for the subworkflow executions.
            wf_2_ex = self._assert_single_item(wf_execs, name='wb.wf2')

            wf_2_task_execs = wf_2_ex.task_executions

            wf_2_task_1_ex = self._assert_single_item(
                wf_2_ex.task_executions,
                name='task1'
            )

            wf_2_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_task_1_ex.id
            )

            wf_3_ex = self._assert_single_item(wf_execs, name='wb.wf3')

            wf_3_task_execs = wf_3_ex.task_executions

            wf_3_task_1_ex = self._assert_single_item(
                wf_3_ex.task_executions,
                name='task1'
            )

            wf_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_3_task_1_ex.id
            )

        self.assertEqual(states.SUCCESS, wf_2_ex.state)
        self.assertEqual(states.SUCCESS, wf_2_task_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_2_task_1_action_exs[0].state)
        self.assertEqual(states.PAUSED, wf_3_ex.state)
        self.assertEqual(states.RUNNING, wf_3_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_3_task_1_action_exs[0].state)
        self.assertEqual(states.SUCCESS, wf_1_task_1_action_exs[0].state)
        self.assertEqual(states.SUCCESS, wf_1_task_1_ex.state)
        self.assertEqual(states.PAUSED, wf_1_task_2_action_exs[0].state)
        self.assertEqual(states.PAUSED, wf_1_task_2_ex.state)
        self.assertEqual(states.PAUSED, wf_1_ex.state)

        # Resume the 2nd subworkflow.
        self.engine.resume_workflow(wf_3_ex.id)

        self.await_workflow_running(wf_3_ex.id)

        with db_api.transaction():
            wf_execs = db_api.get_workflow_executions()

            # Get objects for the parent workflow execution.
            wf_1_ex = self._assert_single_item(wf_execs, name='wb.wf1')

            wf_1_task_execs = wf_1_ex.task_executions

            wf_1_task_1_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task1'
            )

            wf_1_task_1_action_exs = wf_1_task_1_ex.executions

            wf_1_task_2_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task2'
            )

            wf_1_task_2_action_exs = wf_1_task_2_ex.executions

            # Get objects for the subworkflow executions.
            wf_2_ex = self._assert_single_item(wf_execs, name='wb.wf2')

            wf_2_task_execs = wf_2_ex.task_executions

            wf_2_task_1_ex = self._assert_single_item(
                wf_2_ex.task_executions,
                name='task1'
            )

            wf_2_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_task_1_ex.id
            )

            wf_3_ex = self._assert_single_item(wf_execs, name='wb.wf3')

            wf_3_task_execs = wf_3_ex.task_executions

            wf_3_task_1_ex = self._assert_single_item(
                wf_3_ex.task_executions,
                name='task1'
            )

            wf_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_3_task_1_ex.id
            )

        self.assertEqual(states.SUCCESS, wf_2_ex.state)
        self.assertEqual(states.SUCCESS, wf_2_task_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_2_task_1_action_exs[0].state)
        self.assertEqual(states.RUNNING, wf_3_ex.state)
        self.assertEqual(states.RUNNING, wf_3_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_3_task_1_action_exs[0].state)
        self.assertEqual(states.SUCCESS, wf_1_task_1_action_exs[0].state)
        self.assertEqual(states.SUCCESS, wf_1_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_1_task_2_action_exs[0].state)
        self.assertEqual(states.RUNNING, wf_1_task_2_ex.state)
        self.assertEqual(states.RUNNING, wf_1_ex.state)

        # Complete action execution of 2nd subworkflow.
        self.engine.on_action_complete(
            wf_3_task_1_action_exs[0].id,
            ml_actions.Result(data={'result': 'foobar'})
        )

        self.await_workflow_success(wf_3_ex.id)
        self.await_workflow_success(wf_1_ex.id)

        with db_api.transaction():
            wf_execs = db_api.get_workflow_executions()

            # Get objects for the parent workflow execution.
            wf_1_ex = self._assert_single_item(wf_execs, name='wb.wf1')

            wf_1_task_execs = wf_1_ex.task_executions

            wf_1_task_1_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task1'
            )

            wf_1_task_1_action_exs = wf_1_task_1_ex.executions

            wf_1_task_2_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task2'
            )

            wf_1_task_2_action_exs = wf_1_task_2_ex.executions

            wf_1_task_3_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task3'
            )

            # Get objects for the subworkflow executions.
            wf_2_ex = self._assert_single_item(wf_execs, name='wb.wf2')

            wf_2_task_execs = wf_2_ex.task_executions

            wf_2_task_1_ex = self._assert_single_item(
                wf_2_ex.task_executions,
                name='task1'
            )

            wf_2_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_task_1_ex.id
            )

            wf_2_task_2_ex = self._assert_single_item(
                wf_2_ex.task_executions,
                name='task2'
            )

            wf_3_ex = self._assert_single_item(wf_execs, name='wb.wf3')

            wf_3_task_execs = wf_3_ex.task_executions

            wf_3_task_1_ex = self._assert_single_item(
                wf_3_ex.task_executions,
                name='task1'
            )

            wf_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_3_task_1_ex.id
            )

            wf_3_task_2_ex = self._assert_single_item(
                wf_3_ex.task_executions,
                name='task2'
            )

        self.assertEqual(states.SUCCESS, wf_1_ex.state)
        self.assertEqual(3, len(wf_1_task_execs))
        self.assertEqual(states.SUCCESS, wf_1_task_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_1_task_2_ex.state)
        self.assertEqual(states.SUCCESS, wf_1_task_3_ex.state)
        self.assertEqual(states.SUCCESS, wf_1_task_1_action_exs[0].state)
        self.assertEqual(states.SUCCESS, wf_1_task_2_action_exs[0].state)
        self.assertEqual(states.SUCCESS, wf_2_ex.state)
        self.assertEqual(2, len(wf_2_task_execs))
        self.assertEqual(states.SUCCESS, wf_2_task_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_2_task_2_ex.state)
        self.assertEqual(states.SUCCESS, wf_3_ex.state)
        self.assertEqual(2, len(wf_3_task_execs))
        self.assertEqual(states.SUCCESS, wf_3_task_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_3_task_2_ex.state)

    def test_pause_resume_cascade_down_to_with_items_subworkflows(self):
        workbook = """
        version: '2.0'
        name: wb
        workflows:
          wf1:
            tasks:
              task1:
                with-items: i in <% range(3) %>
                workflow: wf2
                on-success:
                  - task3
              task2:
                workflow: wf3
                on-success:
                  - task3
              task3:
                join: all
                action: std.noop
          wf2:
            tasks:
              task1:
                action: std.async_noop
                on-success:
                  - task2
              task2:
                action: std.noop
          wf3:
            tasks:
              task1:
                action: std.async_noop
                on-success:
                  - task2
              task2:
                action: std.noop
        """

        wb_service.create_workbook_v2(workbook)

        # Start workflow execution.
        wf_1_ex = self.engine.start_workflow('wb.wf1')

        self.await_workflow_state(wf_1_ex.id, states.RUNNING)

        with db_api.transaction():
            wf_execs = db_api.get_workflow_executions()

            # Get objects for the parent workflow execution.
            wf_1_ex = self._assert_single_item(wf_execs, name='wb.wf1')

            wf_1_task_execs = wf_1_ex.task_executions

            wf_1_task_1_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task1'
            )

            wf_1_task_1_action_exs = sorted(
                wf_1_task_1_ex.executions,
                key=lambda x: x['runtime_context']['index']
            )

            wf_1_task_2_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task2'
            )

            wf_1_task_2_action_exs = wf_1_task_2_ex.executions

            # Get objects for the with-items subworkflow executions.
            wf_2_ex_1 = db_api.get_workflow_execution(
                wf_1_task_1_action_exs[0].id
            )

            wf_2_ex_1_task_execs = wf_2_ex_1.task_executions

            wf_2_ex_1_task_1_ex = self._assert_single_item(
                wf_2_ex_1.task_executions,
                name='task1'
            )

            wf_2_ex_1_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_ex_1_task_1_ex.id
            )

            wf_2_ex_2 = db_api.get_workflow_execution(
                wf_1_task_1_action_exs[1].id
            )

            wf_2_ex_2_task_execs = wf_2_ex_2.task_executions

            wf_2_ex_2_task_1_ex = self._assert_single_item(
                wf_2_ex_2.task_executions,
                name='task1'
            )

            wf_2_ex_2_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_ex_2_task_1_ex.id
            )

            wf_2_ex_3 = db_api.get_workflow_execution(
                wf_1_task_1_action_exs[2].id
            )

            wf_2_ex_3_task_execs = wf_2_ex_3.task_executions

            wf_2_ex_3_task_1_ex = self._assert_single_item(
                wf_2_ex_3.task_executions,
                name='task1'
            )

            wf_2_ex_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_ex_3_task_1_ex.id
            )

            # Get objects for the wf3 subworkflow execution.
            wf_3_ex = self._assert_single_item(wf_execs, name='wb.wf3')

            wf_3_task_execs = wf_3_ex.task_executions

            wf_3_task_1_ex = self._assert_single_item(
                wf_3_ex.task_executions,
                name='task1'
            )

            wf_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_3_task_1_ex.id
            )

        # Check state of parent workflow execution.
        self.assertEqual(states.RUNNING, wf_1_ex.state)
        self.assertEqual(2, len(wf_1_task_execs))
        self.assertEqual(states.RUNNING, wf_1_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_1_task_2_ex.state)
        self.assertEqual(3, len(wf_1_task_1_action_exs))

        # Check state of wf2 (1) subworkflow execution.
        self.assertEqual(states.RUNNING, wf_1_task_1_action_exs[0].state)
        self.assertEqual(wf_1_task_1_action_exs[0].id, wf_2_ex_1.id)
        self.assertEqual(states.RUNNING, wf_2_ex_1.state)
        self.assertEqual(1, len(wf_2_ex_1_task_execs))
        self.assertEqual(states.RUNNING, wf_2_ex_1_task_1_ex.state)
        self.assertEqual(1, len(wf_2_ex_1_task_1_action_exs))
        self.assertEqual(states.RUNNING, wf_2_ex_1_task_1_action_exs[0].state)

        # Check state of wf2 (2) subworkflow execution.
        self.assertEqual(states.RUNNING, wf_1_task_1_action_exs[1].state)
        self.assertEqual(wf_1_task_1_action_exs[1].id, wf_2_ex_2.id)
        self.assertEqual(states.RUNNING, wf_2_ex_2.state)
        self.assertEqual(1, len(wf_2_ex_2_task_execs))
        self.assertEqual(states.RUNNING, wf_2_ex_2_task_1_ex.state)
        self.assertEqual(1, len(wf_2_ex_2_task_1_action_exs))
        self.assertEqual(states.RUNNING, wf_2_ex_2_task_1_action_exs[0].state)

        # Check state of wf2 (3) subworkflow execution.
        self.assertEqual(states.RUNNING, wf_1_task_1_action_exs[2].state)
        self.assertEqual(wf_1_task_1_action_exs[2].id, wf_2_ex_3.id)
        self.assertEqual(states.RUNNING, wf_2_ex_3.state)
        self.assertEqual(1, len(wf_2_ex_3_task_execs))
        self.assertEqual(states.RUNNING, wf_2_ex_3_task_1_ex.state)
        self.assertEqual(1, len(wf_2_ex_3_task_1_action_exs))
        self.assertEqual(states.RUNNING, wf_2_ex_3_task_1_action_exs[0].state)

        # Check state of wf3 subworkflow execution.
        self.assertEqual(1, len(wf_1_task_2_action_exs))
        self.assertEqual(states.RUNNING, wf_1_task_2_action_exs[0].state)
        self.assertEqual(wf_1_task_2_action_exs[0].id, wf_3_ex.id)
        self.assertEqual(states.RUNNING, wf_3_ex.state)
        self.assertEqual(1, len(wf_3_task_execs))
        self.assertEqual(states.RUNNING, wf_3_task_1_ex.state)
        self.assertEqual(1, len(wf_3_task_1_action_exs))
        self.assertEqual(states.RUNNING, wf_3_task_1_action_exs[0].state)

        # Pause the main workflow.
        self.engine.pause_workflow(wf_1_ex.id)
        self.await_workflow_paused(wf_2_ex_1.id)
        self.await_workflow_paused(wf_2_ex_2.id)
        self.await_workflow_paused(wf_2_ex_3.id)
        self.await_workflow_paused(wf_3_ex.id)
        self.await_task_paused(wf_1_task_1_ex.id)
        self.await_task_paused(wf_1_task_2_ex.id)
        self.await_workflow_paused(wf_1_ex.id)

        with db_api.transaction():
            wf_execs = db_api.get_workflow_executions()

            # Get objects for the parent workflow execution.
            wf_1_ex = self._assert_single_item(wf_execs, name='wb.wf1')

            wf_1_task_execs = wf_1_ex.task_executions

            wf_1_task_1_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task1'
            )

            wf_1_task_1_action_exs = sorted(
                wf_1_task_1_ex.executions,
                key=lambda x: x['runtime_context']['index']
            )

            wf_1_task_2_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task2'
            )

            wf_1_task_2_action_exs = wf_1_task_2_ex.executions

            # Get objects for the with-items subworkflow executions.
            wf_2_ex_1 = db_api.get_workflow_execution(
                wf_1_task_1_action_exs[0].id
            )

            wf_2_ex_1_task_execs = wf_2_ex_1.task_executions

            wf_2_ex_1_task_1_ex = self._assert_single_item(
                wf_2_ex_1.task_executions,
                name='task1'
            )

            wf_2_ex_1_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_ex_1_task_1_ex.id
            )

            wf_2_ex_2 = db_api.get_workflow_execution(
                wf_1_task_1_action_exs[1].id
            )

            wf_2_ex_2_task_execs = wf_2_ex_2.task_executions

            wf_2_ex_2_task_1_ex = self._assert_single_item(
                wf_2_ex_2.task_executions,
                name='task1'
            )

            wf_2_ex_2_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_ex_2_task_1_ex.id
            )

            wf_2_ex_3 = db_api.get_workflow_execution(
                wf_1_task_1_action_exs[2].id
            )

            wf_2_ex_3_task_execs = wf_2_ex_3.task_executions

            wf_2_ex_3_task_1_ex = self._assert_single_item(
                wf_2_ex_3.task_executions,
                name='task1'
            )

            wf_2_ex_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_ex_3_task_1_ex.id
            )

            # Get objects for the wf3 subworkflow execution.
            wf_3_ex = self._assert_single_item(wf_execs, name='wb.wf3')

            wf_3_task_execs = wf_3_ex.task_executions

            wf_3_task_1_ex = self._assert_single_item(
                wf_3_ex.task_executions,
                name='task1'
            )

            wf_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_3_task_1_ex.id
            )

        # Check state of parent workflow execution.
        self.assertEqual(states.PAUSED, wf_1_ex.state)
        self.assertEqual(states.PAUSED, wf_1_task_1_ex.state)
        self.assertEqual(states.PAUSED, wf_1_task_2_ex.state)

        # Check state of wf2 (1) subworkflow execution.
        self.assertEqual(states.PAUSED, wf_1_task_1_action_exs[0].state)
        self.assertEqual(states.PAUSED, wf_2_ex_1.state)
        self.assertEqual(states.RUNNING, wf_2_ex_1_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_2_ex_1_task_1_action_exs[0].state)

        # Check state of wf2 (2) subworkflow execution.
        self.assertEqual(states.PAUSED, wf_1_task_1_action_exs[1].state)
        self.assertEqual(states.PAUSED, wf_2_ex_2.state)
        self.assertEqual(states.RUNNING, wf_2_ex_2_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_2_ex_2_task_1_action_exs[0].state)

        # Check state of wf2 (3) subworkflow execution.
        self.assertEqual(states.PAUSED, wf_1_task_1_action_exs[2].state)
        self.assertEqual(states.PAUSED, wf_2_ex_3.state)
        self.assertEqual(states.RUNNING, wf_2_ex_3_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_2_ex_3_task_1_action_exs[0].state)

        # Check state of wf3 subworkflow execution.
        self.assertEqual(states.PAUSED, wf_1_task_2_action_exs[0].state)
        self.assertEqual(states.PAUSED, wf_3_ex.state)
        self.assertEqual(states.RUNNING, wf_3_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_3_task_1_action_exs[0].state)

        # Resume the main workflow.
        self.engine.resume_workflow(wf_1_ex.id)
        self.await_workflow_running(wf_2_ex_1.id)
        self.await_workflow_running(wf_2_ex_2.id)
        self.await_workflow_running(wf_2_ex_3.id)
        self.await_workflow_running(wf_3_ex.id)
        self.await_task_running(wf_1_task_1_ex.id)
        self.await_task_running(wf_1_task_2_ex.id)
        self.await_workflow_running(wf_1_ex.id)

        with db_api.transaction():
            wf_execs = db_api.get_workflow_executions()

            # Get objects for the parent workflow execution.
            wf_1_ex = self._assert_single_item(wf_execs, name='wb.wf1')

            wf_1_task_execs = wf_1_ex.task_executions

            wf_1_task_1_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task1'
            )

            wf_1_task_1_action_exs = sorted(
                wf_1_task_1_ex.executions,
                key=lambda x: x['runtime_context']['index']
            )

            wf_1_task_2_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task2'
            )

            wf_1_task_2_action_exs = wf_1_task_2_ex.executions

            # Get objects for the with-items subworkflow executions.
            wf_2_ex_1 = db_api.get_workflow_execution(
                wf_1_task_1_action_exs[0].id
            )

            wf_2_ex_1_task_execs = wf_2_ex_1.task_executions

            wf_2_ex_1_task_1_ex = self._assert_single_item(
                wf_2_ex_1.task_executions,
                name='task1'
            )

            wf_2_ex_1_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_ex_1_task_1_ex.id
            )

            wf_2_ex_2 = db_api.get_workflow_execution(
                wf_1_task_1_action_exs[1].id
            )

            wf_2_ex_2_task_execs = wf_2_ex_2.task_executions

            wf_2_ex_2_task_1_ex = self._assert_single_item(
                wf_2_ex_2.task_executions,
                name='task1'
            )

            wf_2_ex_2_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_ex_2_task_1_ex.id
            )

            wf_2_ex_3 = db_api.get_workflow_execution(
                wf_1_task_1_action_exs[2].id
            )

            wf_2_ex_3_task_execs = wf_2_ex_3.task_executions

            wf_2_ex_3_task_1_ex = self._assert_single_item(
                wf_2_ex_3.task_executions,
                name='task1'
            )

            wf_2_ex_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_ex_3_task_1_ex.id
            )

            # Get objects for the wf3 subworkflow execution.
            wf_3_ex = self._assert_single_item(wf_execs, name='wb.wf3')

            wf_3_task_execs = wf_3_ex.task_executions

            wf_3_task_1_ex = self._assert_single_item(
                wf_3_ex.task_executions,
                name='task1'
            )

            wf_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_3_task_1_ex.id
            )

        # Check state of parent workflow execution.
        self.assertEqual(states.RUNNING, wf_1_ex.state)
        self.assertEqual(states.RUNNING, wf_1_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_1_task_2_ex.state)

        # Check state of wf2 (1) subworkflow execution.
        self.assertEqual(states.RUNNING, wf_1_task_1_action_exs[0].state)
        self.assertEqual(states.RUNNING, wf_2_ex_1.state)
        self.assertEqual(states.RUNNING, wf_2_ex_1_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_2_ex_1_task_1_action_exs[0].state)

        # Check state of wf2 (2) subworkflow execution.
        self.assertEqual(states.RUNNING, wf_1_task_1_action_exs[1].state)
        self.assertEqual(states.RUNNING, wf_2_ex_2.state)
        self.assertEqual(states.RUNNING, wf_2_ex_2_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_2_ex_2_task_1_action_exs[0].state)

        # Check state of wf2 (3) subworkflow execution.
        self.assertEqual(states.RUNNING, wf_1_task_1_action_exs[2].state)
        self.assertEqual(states.RUNNING, wf_2_ex_3.state)
        self.assertEqual(states.RUNNING, wf_2_ex_3_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_2_ex_3_task_1_action_exs[0].state)

        # Check state of wf3 subworkflow execution.
        self.assertEqual(states.RUNNING, wf_1_task_2_action_exs[0].state)
        self.assertEqual(states.RUNNING, wf_3_ex.state)
        self.assertEqual(states.RUNNING, wf_3_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_3_task_1_action_exs[0].state)

        # Complete action execution of subworkflows.
        self.engine.on_action_complete(
            wf_2_ex_1_task_1_action_exs[0].id,
            ml_actions.Result(data={'result': 'foobar'})
        )

        self.engine.on_action_complete(
            wf_2_ex_2_task_1_action_exs[0].id,
            ml_actions.Result(data={'result': 'foobar'})
        )

        self.engine.on_action_complete(
            wf_2_ex_3_task_1_action_exs[0].id,
            ml_actions.Result(data={'result': 'foobar'})
        )

        self.engine.on_action_complete(
            wf_3_task_1_action_exs[0].id,
            ml_actions.Result(data={'result': 'foobar'})
        )

        self.await_workflow_success(wf_2_ex_1.id)
        self.await_workflow_success(wf_2_ex_2.id)
        self.await_workflow_success(wf_2_ex_3.id)
        self.await_workflow_success(wf_3_ex.id)
        self.await_task_success(wf_1_task_1_ex.id)
        self.await_task_success(wf_1_task_2_ex.id)
        self.await_workflow_success(wf_1_ex.id)

        with db_api.transaction():
            wf_execs = db_api.get_workflow_executions()

            # Get objects for the parent workflow execution.
            wf_1_ex = self._assert_single_item(wf_execs, name='wb.wf1')

            wf_1_task_execs = wf_1_ex.task_executions

            wf_1_task_1_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task1'
            )

            wf_1_task_1_action_exs = sorted(
                wf_1_task_1_ex.executions,
                key=lambda x: x['runtime_context']['index']
            )

            wf_1_task_2_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task2'
            )

            wf_1_task_2_action_exs = wf_1_task_2_ex.executions

            # Get objects for the with-items subworkflow executions.
            wf_2_ex_1 = db_api.get_workflow_execution(
                wf_1_task_1_action_exs[0].id
            )

            wf_2_ex_1_task_execs = wf_2_ex_1.task_executions

            wf_2_ex_1_task_1_ex = self._assert_single_item(
                wf_2_ex_1.task_executions,
                name='task1'
            )

            wf_2_ex_1_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_ex_1_task_1_ex.id
            )

            wf_2_ex_2 = db_api.get_workflow_execution(
                wf_1_task_1_action_exs[1].id
            )

            wf_2_ex_2_task_execs = wf_2_ex_2.task_executions

            wf_2_ex_2_task_1_ex = self._assert_single_item(
                wf_2_ex_2.task_executions,
                name='task1'
            )

            wf_2_ex_2_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_ex_2_task_1_ex.id
            )

            wf_2_ex_3 = db_api.get_workflow_execution(
                wf_1_task_1_action_exs[2].id
            )

            wf_2_ex_3_task_execs = wf_2_ex_3.task_executions

            wf_2_ex_3_task_1_ex = self._assert_single_item(
                wf_2_ex_3.task_executions,
                name='task1'
            )

            wf_2_ex_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_ex_3_task_1_ex.id
            )

            # Get objects for the wf3 subworkflow execution.
            wf_3_ex = self._assert_single_item(wf_execs, name='wb.wf3')

            wf_3_task_execs = wf_3_ex.task_executions

            wf_3_task_1_ex = self._assert_single_item(
                wf_3_ex.task_executions,
                name='task1'
            )

            wf_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_3_task_1_ex.id
            )

        # Check state of parent workflow execution.
        self.assertEqual(states.SUCCESS, wf_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_1_task_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_1_task_2_ex.state)

        # Check state of wf2 (1) subworkflow execution.
        self.assertEqual(states.SUCCESS, wf_1_task_1_action_exs[0].state)
        self.assertEqual(states.SUCCESS, wf_2_ex_1.state)
        self.assertEqual(states.SUCCESS, wf_2_ex_1_task_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_2_ex_1_task_1_action_exs[0].state)

        # Check state of wf2 (2) subworkflow execution.
        self.assertEqual(states.SUCCESS, wf_1_task_1_action_exs[1].state)
        self.assertEqual(states.SUCCESS, wf_2_ex_2.state)
        self.assertEqual(states.SUCCESS, wf_2_ex_2_task_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_2_ex_2_task_1_action_exs[0].state)

        # Check state of wf2 (3) subworkflow execution.
        self.assertEqual(states.SUCCESS, wf_1_task_1_action_exs[2].state)
        self.assertEqual(states.SUCCESS, wf_2_ex_3.state)
        self.assertEqual(states.SUCCESS, wf_2_ex_3_task_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_2_ex_3_task_1_action_exs[0].state)

        # Check state of wf3 subworkflow execution.
        self.assertEqual(states.SUCCESS, wf_1_task_2_action_exs[0].state)
        self.assertEqual(states.SUCCESS, wf_3_ex.state)
        self.assertEqual(states.SUCCESS, wf_3_task_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_3_task_1_action_exs[0].state)

    def test_pause_resume_cascade_up_from_with_items_subworkflow(self):
        workbook = """
        version: '2.0'
        name: wb
        workflows:
          wf1:
            tasks:
              task1:
                with-items: i in <% range(3) %>
                workflow: wf2
                on-success:
                  - task3
              task2:
                workflow: wf3
                on-success:
                  - task3
              task3:
                join: all
                action: std.noop
          wf2:
            tasks:
              task1:
                action: std.async_noop
                on-success:
                  - task2
              task2:
                action: std.noop
          wf3:
            tasks:
              task1:
                action: std.async_noop
                on-success:
                  - task2
              task2:
                action: std.noop
        """

        wb_service.create_workbook_v2(workbook)

        # Start workflow execution.
        wf_1_ex = self.engine.start_workflow('wb.wf1')

        self.await_workflow_state(wf_1_ex.id, states.RUNNING)

        with db_api.transaction():
            wf_execs = db_api.get_workflow_executions()

            # Get objects for the parent workflow execution.
            wf_1_ex = self._assert_single_item(wf_execs, name='wb.wf1')

            wf_1_task_execs = wf_1_ex.task_executions

            wf_1_task_1_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task1'
            )

            wf_1_task_1_action_exs = sorted(
                wf_1_task_1_ex.executions,
                key=lambda x: x['runtime_context']['index']
            )

            wf_1_task_2_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task2'
            )

            wf_1_task_2_action_exs = wf_1_task_2_ex.executions

            # Get objects for the with-items subworkflow executions.
            wf_2_ex_1 = db_api.get_workflow_execution(
                wf_1_task_1_action_exs[0].id
            )

            wf_2_ex_1_task_execs = wf_2_ex_1.task_executions

            wf_2_ex_1_task_1_ex = self._assert_single_item(
                wf_2_ex_1.task_executions,
                name='task1'
            )

            wf_2_ex_1_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_ex_1_task_1_ex.id
            )

            wf_2_ex_2 = db_api.get_workflow_execution(
                wf_1_task_1_action_exs[1].id
            )

            wf_2_ex_2_task_execs = wf_2_ex_2.task_executions

            wf_2_ex_2_task_1_ex = self._assert_single_item(
                wf_2_ex_2.task_executions,
                name='task1'
            )

            wf_2_ex_2_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_ex_2_task_1_ex.id
            )

            wf_2_ex_3 = db_api.get_workflow_execution(
                wf_1_task_1_action_exs[2].id
            )

            wf_2_ex_3_task_execs = wf_2_ex_3.task_executions

            wf_2_ex_3_task_1_ex = self._assert_single_item(
                wf_2_ex_3.task_executions,
                name='task1'
            )

            wf_2_ex_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_ex_3_task_1_ex.id
            )

            # Get objects for the wf3 subworkflow execution.
            wf_3_ex = self._assert_single_item(wf_execs, name='wb.wf3')

            wf_3_task_execs = wf_3_ex.task_executions

            wf_3_task_1_ex = self._assert_single_item(
                wf_3_ex.task_executions,
                name='task1'
            )

            wf_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_3_task_1_ex.id
            )

        # Check state of parent workflow execution.
        self.assertEqual(states.RUNNING, wf_1_ex.state)
        self.assertEqual(2, len(wf_1_task_execs))
        self.assertEqual(states.RUNNING, wf_1_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_1_task_2_ex.state)
        self.assertEqual(3, len(wf_1_task_1_action_exs))

        # Check state of wf2 (1) subworkflow execution.
        self.assertEqual(states.RUNNING, wf_1_task_1_action_exs[0].state)
        self.assertEqual(wf_1_task_1_action_exs[0].id, wf_2_ex_1.id)
        self.assertEqual(states.RUNNING, wf_2_ex_1.state)
        self.assertEqual(1, len(wf_2_ex_1_task_execs))
        self.assertEqual(states.RUNNING, wf_2_ex_1_task_1_ex.state)
        self.assertEqual(1, len(wf_2_ex_1_task_1_action_exs))
        self.assertEqual(states.RUNNING, wf_2_ex_1_task_1_action_exs[0].state)

        # Check state of wf2 (2) subworkflow execution.
        self.assertEqual(states.RUNNING, wf_1_task_1_action_exs[1].state)
        self.assertEqual(wf_1_task_1_action_exs[1].id, wf_2_ex_2.id)
        self.assertEqual(states.RUNNING, wf_2_ex_2.state)
        self.assertEqual(1, len(wf_2_ex_2_task_execs))
        self.assertEqual(states.RUNNING, wf_2_ex_2_task_1_ex.state)
        self.assertEqual(1, len(wf_2_ex_2_task_1_action_exs))
        self.assertEqual(states.RUNNING, wf_2_ex_2_task_1_action_exs[0].state)

        # Check state of wf2 (3) subworkflow execution.
        self.assertEqual(states.RUNNING, wf_1_task_1_action_exs[2].state)
        self.assertEqual(wf_1_task_1_action_exs[2].id, wf_2_ex_3.id)
        self.assertEqual(states.RUNNING, wf_2_ex_3.state)
        self.assertEqual(1, len(wf_2_ex_3_task_execs))
        self.assertEqual(states.RUNNING, wf_2_ex_3_task_1_ex.state)
        self.assertEqual(1, len(wf_2_ex_3_task_1_action_exs))
        self.assertEqual(states.RUNNING, wf_2_ex_3_task_1_action_exs[0].state)

        # Check state of wf3 subworkflow execution.
        self.assertEqual(1, len(wf_1_task_2_action_exs))
        self.assertEqual(states.RUNNING, wf_1_task_2_action_exs[0].state)
        self.assertEqual(wf_1_task_2_action_exs[0].id, wf_3_ex.id)
        self.assertEqual(states.RUNNING, wf_3_ex.state)
        self.assertEqual(1, len(wf_3_task_execs))
        self.assertEqual(states.RUNNING, wf_3_task_1_ex.state)
        self.assertEqual(1, len(wf_3_task_1_action_exs))
        self.assertEqual(states.RUNNING, wf_3_task_1_action_exs[0].state)

        # Pause one of the subworkflows in the with-items task.
        self.engine.pause_workflow(wf_2_ex_1.id)

        self.await_workflow_paused(wf_2_ex_1.id)
        self.await_workflow_paused(wf_2_ex_2.id)
        self.await_workflow_paused(wf_2_ex_3.id)
        self.await_workflow_paused(wf_3_ex.id)
        self.await_task_paused(wf_1_task_1_ex.id)
        self.await_task_paused(wf_1_task_2_ex.id)
        self.await_workflow_paused(wf_1_ex.id)

        with db_api.transaction():
            wf_execs = db_api.get_workflow_executions()

            # Get objects for the parent workflow execution.
            wf_1_ex = self._assert_single_item(wf_execs, name='wb.wf1')

            wf_1_task_1_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task1'
            )

            wf_1_task_1_action_exs = sorted(
                wf_1_task_1_ex.executions,
                key=lambda x: x['runtime_context']['index']
            )

            wf_1_task_2_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task2'
            )

            wf_1_task_2_action_exs = wf_1_task_2_ex.executions

            # Get objects for the with-items subworkflow executions.
            wf_2_ex_1 = db_api.get_workflow_execution(
                wf_1_task_1_action_exs[0].id
            )

            wf_2_ex_1_task_1_ex = self._assert_single_item(
                wf_2_ex_1.task_executions,
                name='task1'
            )

            wf_2_ex_1_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_ex_1_task_1_ex.id
            )

            wf_2_ex_2 = db_api.get_workflow_execution(
                wf_1_task_1_action_exs[1].id
            )

            wf_2_ex_2_task_1_ex = self._assert_single_item(
                wf_2_ex_2.task_executions,
                name='task1'
            )

            wf_2_ex_2_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_ex_2_task_1_ex.id
            )

            wf_2_ex_3 = db_api.get_workflow_execution(
                wf_1_task_1_action_exs[2].id
            )

            wf_2_ex_3_task_1_ex = self._assert_single_item(
                wf_2_ex_3.task_executions,
                name='task1'
            )

            wf_2_ex_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_ex_3_task_1_ex.id
            )

            # Get objects for the wf3 subworkflow execution.
            wf_3_ex = self._assert_single_item(wf_execs, name='wb.wf3')

            wf_3_task_1_ex = self._assert_single_item(
                wf_3_ex.task_executions,
                name='task1'
            )

            wf_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_3_task_1_ex.id
            )

        # Check state of parent workflow execution.
        self.assertEqual(states.PAUSED, wf_1_ex.state)
        self.assertEqual(states.PAUSED, wf_1_task_1_ex.state)
        self.assertEqual(states.PAUSED, wf_1_task_2_ex.state)

        # Check state of wf2 (1) subworkflow execution.
        self.assertEqual(states.PAUSED, wf_1_task_1_action_exs[0].state)
        self.assertEqual(states.PAUSED, wf_2_ex_1.state)
        self.assertEqual(states.RUNNING, wf_2_ex_1_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_2_ex_1_task_1_action_exs[0].state)

        # Check state of wf2 (2) subworkflow execution.
        self.assertEqual(states.PAUSED, wf_1_task_1_action_exs[1].state)
        self.assertEqual(states.PAUSED, wf_2_ex_2.state)
        self.assertEqual(states.RUNNING, wf_2_ex_2_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_2_ex_2_task_1_action_exs[0].state)

        # Check state of wf2 (3) subworkflow execution.
        self.assertEqual(states.PAUSED, wf_1_task_1_action_exs[2].state)
        self.assertEqual(states.PAUSED, wf_2_ex_3.state)
        self.assertEqual(states.RUNNING, wf_2_ex_3_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_2_ex_3_task_1_action_exs[0].state)

        # Check state of wf3 subworkflow execution.
        self.assertEqual(states.PAUSED, wf_1_task_2_action_exs[0].state)
        self.assertEqual(states.PAUSED, wf_3_ex.state)
        self.assertEqual(states.RUNNING, wf_3_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_3_task_1_action_exs[0].state)

        # NOTE(rakhmerov): Since cascade pausing is not atomic we need
        # to make sure that all internal operations related to pausing
        # one of workflow executions 'wb.wf2' are completed. So we have
        # to look if any "_on_action_update" calls are scheduled.

        def _predicate():
            return all(
                [
                    '_on_action_update' not in c.target_method_name
                    for c in db_api.get_delayed_calls()
                ]
            )

        self._await(_predicate)

        # Resume one of the subworkflows in the with-items task.
        self.engine.resume_workflow(wf_2_ex_1.id)

        self.await_workflow_running(wf_2_ex_1.id)
        self.await_workflow_paused(wf_2_ex_2.id)
        self.await_workflow_paused(wf_2_ex_3.id)
        self.await_workflow_paused(wf_3_ex.id)
        self.await_task_paused(wf_1_task_1_ex.id)
        self.await_task_paused(wf_1_task_2_ex.id)
        self.await_workflow_paused(wf_1_ex.id)

        # Complete action execution of the subworkflow that is resumed.
        self.engine.on_action_complete(
            wf_2_ex_1_task_1_action_exs[0].id,
            ml_actions.Result(data={'result': 'foobar'})
        )

        self.await_workflow_success(wf_2_ex_1.id)
        self.await_workflow_paused(wf_2_ex_2.id)
        self.await_workflow_paused(wf_2_ex_3.id)
        self.await_workflow_paused(wf_3_ex.id)
        self.await_task_paused(wf_1_task_1_ex.id)
        self.await_task_paused(wf_1_task_2_ex.id)
        self.await_workflow_paused(wf_1_ex.id)

        with db_api.transaction():
            wf_execs = db_api.get_workflow_executions()

            # Get objects for the parent workflow execution.
            wf_1_ex = self._assert_single_item(wf_execs, name='wb.wf1')

            wf_1_task_1_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task1'
            )

            wf_1_task_1_action_exs = sorted(
                wf_1_task_1_ex.executions,
                key=lambda x: x['runtime_context']['index']
            )

            wf_1_task_2_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task2'
            )

            wf_1_task_2_action_exs = wf_1_task_2_ex.executions

            # Get objects for the with-items subworkflow executions.
            wf_2_ex_1 = db_api.get_workflow_execution(
                wf_1_task_1_action_exs[0].id
            )

            wf_2_ex_1_task_1_ex = self._assert_single_item(
                wf_2_ex_1.task_executions,
                name='task1'
            )

            wf_2_ex_1_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_ex_1_task_1_ex.id
            )

            wf_2_ex_2 = db_api.get_workflow_execution(
                wf_1_task_1_action_exs[1].id
            )

            wf_2_ex_2_task_1_ex = self._assert_single_item(
                wf_2_ex_2.task_executions,
                name='task1'
            )

            wf_2_ex_2_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_ex_2_task_1_ex.id
            )

            wf_2_ex_3 = db_api.get_workflow_execution(
                wf_1_task_1_action_exs[2].id
            )

            wf_2_ex_3_task_execs = wf_2_ex_3.task_executions

            wf_2_ex_3_task_1_ex = self._assert_single_item(
                wf_2_ex_3.task_executions,
                name='task1'
            )

            wf_2_ex_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_ex_3_task_1_ex.id
            )

            # Get objects for the wf3 subworkflow execution.
            wf_3_ex = self._assert_single_item(wf_execs, name='wb.wf3')

            wf_3_task_execs = wf_3_ex.task_executions

            wf_3_task_1_ex = self._assert_single_item(
                wf_3_ex.task_executions,
                name='task1'
            )

            wf_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_3_task_1_ex.id
            )

        # Check state of parent workflow execution.
        self.assertEqual(states.PAUSED, wf_1_ex.state)
        self.assertEqual(states.PAUSED, wf_1_task_1_ex.state)
        self.assertEqual(states.PAUSED, wf_1_task_2_ex.state)

        # Check state of wf2 (1) subworkflow execution.
        self.assertEqual(states.SUCCESS, wf_1_task_1_action_exs[0].state)
        self.assertEqual(states.SUCCESS, wf_2_ex_1.state)
        self.assertEqual(states.SUCCESS, wf_2_ex_1_task_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_2_ex_1_task_1_action_exs[0].state)

        # Check state of wf2 (2) subworkflow execution.
        self.assertEqual(states.PAUSED, wf_1_task_1_action_exs[1].state)
        self.assertEqual(states.PAUSED, wf_2_ex_2.state)
        self.assertEqual(states.RUNNING, wf_2_ex_2_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_2_ex_2_task_1_action_exs[0].state)

        # Check state of wf2 (3) subworkflow execution.
        self.assertEqual(states.PAUSED, wf_1_task_1_action_exs[2].state)
        self.assertEqual(states.PAUSED, wf_2_ex_3.state)
        self.assertEqual(states.RUNNING, wf_2_ex_3_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_2_ex_3_task_1_action_exs[0].state)

        # Check state of wf3 subworkflow execution.
        self.assertEqual(states.PAUSED, wf_1_task_2_action_exs[0].state)
        self.assertEqual(states.PAUSED, wf_3_ex.state)
        self.assertEqual(states.RUNNING, wf_3_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_3_task_1_action_exs[0].state)

        # Resume one of the remaining subworkflows.
        self.engine.resume_workflow(wf_2_ex_2.id)
        self.engine.resume_workflow(wf_2_ex_3.id)
        self.engine.resume_workflow(wf_3_ex.id)
        self.await_workflow_running(wf_2_ex_2.id)
        self.await_workflow_running(wf_2_ex_3.id)
        self.await_workflow_running(wf_3_ex.id)
        self.await_task_running(wf_1_task_1_ex.id)
        self.await_task_running(wf_1_task_2_ex.id)
        self.await_workflow_running(wf_1_ex.id)

        # Complete action executions of the remaining subworkflows.
        self.engine.on_action_complete(
            wf_2_ex_2_task_1_action_exs[0].id,
            ml_actions.Result(data={'result': 'foobar'})
        )

        self.engine.on_action_complete(
            wf_2_ex_3_task_1_action_exs[0].id,
            ml_actions.Result(data={'result': 'foobar'})
        )

        self.engine.on_action_complete(
            wf_3_task_1_action_exs[0].id,
            ml_actions.Result(data={'result': 'foobar'})
        )

        self.await_workflow_success(wf_2_ex_1.id)
        self.await_workflow_success(wf_2_ex_2.id)
        self.await_workflow_success(wf_2_ex_3.id)
        self.await_workflow_success(wf_3_ex.id)
        self.await_task_success(wf_1_task_1_ex.id)
        self.await_task_success(wf_1_task_2_ex.id)
        self.await_workflow_success(wf_1_ex.id)

        with db_api.transaction():
            wf_execs = db_api.get_workflow_executions()

            # Get objects for the parent workflow execution.
            wf_1_ex = self._assert_single_item(wf_execs, name='wb.wf1')

            wf_1_task_execs = wf_1_ex.task_executions

            wf_1_task_1_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task1'
            )

            wf_1_task_1_action_exs = sorted(
                wf_1_task_1_ex.executions,
                key=lambda x: x['runtime_context']['index']
            )

            wf_1_task_2_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task2'
            )

            wf_1_task_2_action_exs = wf_1_task_2_ex.executions

            # Get objects for the with-items subworkflow executions.
            wf_2_ex_1 = db_api.get_workflow_execution(
                wf_1_task_1_action_exs[0].id
            )

            wf_2_ex_1_task_execs = wf_2_ex_1.task_executions

            wf_2_ex_1_task_1_ex = self._assert_single_item(
                wf_2_ex_1.task_executions,
                name='task1'
            )

            wf_2_ex_1_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_ex_1_task_1_ex.id
            )

            wf_2_ex_2 = db_api.get_workflow_execution(
                wf_1_task_1_action_exs[1].id
            )

            wf_2_ex_2_task_execs = wf_2_ex_2.task_executions

            wf_2_ex_2_task_1_ex = self._assert_single_item(
                wf_2_ex_2.task_executions,
                name='task1'
            )

            wf_2_ex_2_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_ex_2_task_1_ex.id
            )

            wf_2_ex_3 = db_api.get_workflow_execution(
                wf_1_task_1_action_exs[2].id
            )

            wf_2_ex_3_task_execs = wf_2_ex_3.task_executions

            wf_2_ex_3_task_1_ex = self._assert_single_item(
                wf_2_ex_3.task_executions,
                name='task1'
            )

            wf_2_ex_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_ex_3_task_1_ex.id
            )

            # Get objects for the wf3 subworkflow execution.
            wf_3_ex = self._assert_single_item(wf_execs, name='wb.wf3')

            wf_3_task_execs = wf_3_ex.task_executions

            wf_3_task_1_ex = self._assert_single_item(
                wf_3_ex.task_executions,
                name='task1'
            )

            wf_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_3_task_1_ex.id
            )

        # Check state of parent workflow execution.
        self.assertEqual(states.SUCCESS, wf_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_1_task_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_1_task_2_ex.state)

        # Check state of wf2 (1) subworkflow execution.
        self.assertEqual(states.SUCCESS, wf_1_task_1_action_exs[0].state)
        self.assertEqual(states.SUCCESS, wf_2_ex_1.state)
        self.assertEqual(states.SUCCESS, wf_2_ex_1_task_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_2_ex_1_task_1_action_exs[0].state)

        # Check state of wf2 (2) subworkflow execution.
        self.assertEqual(states.SUCCESS, wf_1_task_1_action_exs[1].state)
        self.assertEqual(states.SUCCESS, wf_2_ex_2.state)
        self.assertEqual(states.SUCCESS, wf_2_ex_2_task_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_2_ex_2_task_1_action_exs[0].state)

        # Check state of wf2 (3) subworkflow execution.
        self.assertEqual(states.SUCCESS, wf_1_task_1_action_exs[2].state)
        self.assertEqual(states.SUCCESS, wf_2_ex_3.state)
        self.assertEqual(states.SUCCESS, wf_2_ex_3_task_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_2_ex_3_task_1_action_exs[0].state)

        # Check state of wf3 subworkflow execution.
        self.assertEqual(states.SUCCESS, wf_1_task_2_action_exs[0].state)
        self.assertEqual(states.SUCCESS, wf_3_ex.state)
        self.assertEqual(states.SUCCESS, wf_3_task_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_3_task_1_action_exs[0].state)

    def test_pause_resume_cascade_up_from_subworkflow_pause_before(self):
        workbook = """
        version: '2.0'
        name: wb
        workflows:
            wf1:
                tasks:
                    task1:
                        workflow: wf2
                        on-success:
                            - task3
                    task2:
                        workflow: wf3
                        on-success:
                            - task3
                    task3:
                        join: all
                        action: std.noop
            wf2:
                tasks:
                    task1:
                        action: std.noop
                        on-success:
                            - task2
                    task2:
                        pause-before: true
                        action: std.async_noop
            wf3:
                tasks:
                    task1:
                        action: std.async_noop
                        on-success:
                            - task2
                    task2:
                        action: std.noop
        """

        wb_service.create_workbook_v2(workbook)

        # Start workflow execution.
        wf_1_ex = self.engine.start_workflow('wb.wf1')

        self.await_workflow_state(wf_1_ex.id, states.PAUSED)

        with db_api.transaction():
            wf_execs = db_api.get_workflow_executions()

            # Get objects for the parent workflow execution.
            wf_1_ex = self._assert_single_item(wf_execs, name='wb.wf1')

            wf_1_task_execs = wf_1_ex.task_executions

            wf_1_task_1_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task1'
            )

            wf_1_task_1_action_exs = wf_1_task_1_ex.executions

            wf_1_task_2_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task2'
            )

            wf_1_task_2_action_exs = wf_1_task_2_ex.executions

            # Get objects for the subworkflow executions.
            wf_2_ex = self._assert_single_item(wf_execs, name='wb.wf2')

            wf_2_task_execs = wf_2_ex.task_executions

            wf_2_task_1_ex = self._assert_single_item(
                wf_2_ex.task_executions,
                name='task1'
            )

            wf_2_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_task_1_ex.id
            )

            wf_2_task_2_ex = self._assert_single_item(
                wf_2_ex.task_executions,
                name='task2'
            )

            wf_2_task_2_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_task_2_ex.id
            )

            wf_3_ex = self._assert_single_item(wf_execs, name='wb.wf3')

            wf_3_task_execs = wf_3_ex.task_executions

            wf_3_task_1_ex = self._assert_single_item(
                wf_3_ex.task_executions,
                name='task1'
            )

            wf_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_3_task_1_ex.id
            )

        self.assertEqual(states.PAUSED, wf_2_ex.state)
        self.assertEqual(states.SUCCESS, wf_2_task_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_2_task_1_action_exs[0].state)
        self.assertEqual(states.IDLE, wf_2_task_2_ex.state)
        self.assertEqual(0, len(wf_2_task_2_action_exs))
        self.assertEqual(states.PAUSED, wf_3_ex.state)
        self.assertEqual(states.RUNNING, wf_3_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_3_task_1_action_exs[0].state)
        self.assertEqual(states.PAUSED, wf_1_task_1_action_exs[0].state)
        self.assertEqual(states.PAUSED, wf_1_task_1_ex.state)
        self.assertEqual(states.PAUSED, wf_1_task_2_action_exs[0].state)
        self.assertEqual(states.PAUSED, wf_1_task_2_ex.state)
        self.assertEqual(states.PAUSED, wf_1_ex.state)

        # Resume the main workflow.
        self.engine.resume_workflow(wf_1_ex.id)

        self.await_workflow_running(wf_1_ex.id)
        self.await_workflow_running(wf_2_ex.id)
        self.await_workflow_running(wf_3_ex.id)

        with db_api.transaction():
            wf_execs = db_api.get_workflow_executions()

            # Get objects for the parent workflow execution.
            wf_1_ex = self._assert_single_item(wf_execs, name='wb.wf1')

            wf_1_task_execs = wf_1_ex.task_executions

            wf_1_task_1_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task1'
            )

            wf_1_task_1_action_exs = wf_1_task_1_ex.executions

            wf_1_task_2_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task2'
            )

            wf_1_task_2_action_exs = wf_1_task_2_ex.executions

            # Get objects for the subworkflow executions.
            wf_2_ex = self._assert_single_item(wf_execs, name='wb.wf2')

            wf_2_task_execs = wf_2_ex.task_executions

            wf_2_task_1_ex = self._assert_single_item(
                wf_2_ex.task_executions,
                name='task1'
            )

            wf_2_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_task_1_ex.id
            )

            wf_2_task_2_ex = self._assert_single_item(
                wf_2_ex.task_executions,
                name='task2'
            )

            wf_2_task_2_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_task_2_ex.id
            )

            wf_3_ex = self._assert_single_item(wf_execs, name='wb.wf3')

            wf_3_task_execs = wf_3_ex.task_executions

            wf_3_task_1_ex = self._assert_single_item(
                wf_3_ex.task_executions,
                name='task1'
            )

            wf_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_3_task_1_ex.id
            )

        self.assertEqual(states.RUNNING, wf_2_ex.state)
        self.assertEqual(states.SUCCESS, wf_2_task_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_2_task_1_action_exs[0].state)
        self.assertEqual(states.RUNNING, wf_2_task_2_ex.state)
        self.assertEqual(states.RUNNING, wf_2_task_2_action_exs[0].state)
        self.assertEqual(states.RUNNING, wf_3_ex.state)
        self.assertEqual(states.RUNNING, wf_3_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_3_task_1_action_exs[0].state)
        self.assertEqual(states.RUNNING, wf_1_task_1_action_exs[0].state)
        self.assertEqual(states.RUNNING, wf_1_task_1_ex.state)
        self.assertEqual(states.RUNNING, wf_1_task_2_action_exs[0].state)
        self.assertEqual(states.RUNNING, wf_1_task_2_ex.state)
        self.assertEqual(states.RUNNING, wf_1_ex.state)

        # Complete action executions of the subworkflows.
        self.engine.on_action_complete(
            wf_2_task_2_action_exs[0].id,
            ml_actions.Result(data={'result': 'foobar'})
        )

        self.engine.on_action_complete(
            wf_3_task_1_action_exs[0].id,
            ml_actions.Result(data={'result': 'foobar'})
        )

        self.await_workflow_success(wf_2_ex.id)
        self.await_workflow_success(wf_3_ex.id)
        self.await_workflow_success(wf_1_ex.id)

        with db_api.transaction():
            wf_execs = db_api.get_workflow_executions()

            # Get objects for the parent workflow execution.
            wf_1_ex = self._assert_single_item(wf_execs, name='wb.wf1')

            wf_1_task_execs = wf_1_ex.task_executions

            wf_1_task_1_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task1'
            )

            wf_1_task_1_action_exs = wf_1_task_1_ex.executions

            wf_1_task_2_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task2'
            )

            wf_1_task_2_action_exs = wf_1_task_2_ex.executions

            wf_1_task_3_ex = self._assert_single_item(
                wf_1_ex.task_executions,
                name='task3'
            )

            # Get objects for the subworkflow executions.
            wf_2_ex = self._assert_single_item(wf_execs, name='wb.wf2')

            wf_2_task_execs = wf_2_ex.task_executions

            wf_2_task_1_ex = self._assert_single_item(
                wf_2_ex.task_executions,
                name='task1'
            )

            wf_2_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_task_1_ex.id
            )

            wf_2_task_2_ex = self._assert_single_item(
                wf_2_ex.task_executions,
                name='task2'
            )

            wf_2_task_2_action_exs = db_api.get_action_executions(
                task_execution_id=wf_2_task_2_ex.id
            )

            wf_3_ex = self._assert_single_item(wf_execs, name='wb.wf3')

            wf_3_task_execs = wf_3_ex.task_executions

            wf_3_task_1_ex = self._assert_single_item(
                wf_3_ex.task_executions,
                name='task1'
            )

            wf_3_task_1_action_exs = db_api.get_action_executions(
                task_execution_id=wf_3_task_1_ex.id
            )

            wf_3_task_2_ex = self._assert_single_item(
                wf_3_ex.task_executions,
                name='task2'
            )

            wf_3_task_2_action_exs = db_api.get_action_executions(
                task_execution_id=wf_3_task_2_ex.id
            )

        self.assertEqual(states.SUCCESS, wf_1_ex.state)
        self.assertEqual(3, len(wf_1_task_execs))
        self.assertEqual(states.SUCCESS, wf_1_task_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_1_task_2_ex.state)
        self.assertEqual(states.SUCCESS, wf_1_task_3_ex.state)
        self.assertEqual(states.SUCCESS, wf_1_task_1_action_exs[0].state)
        self.assertEqual(states.SUCCESS, wf_1_task_2_action_exs[0].state)
        self.assertEqual(states.SUCCESS, wf_2_ex.state)
        self.assertEqual(2, len(wf_2_task_execs))
        self.assertEqual(states.SUCCESS, wf_2_task_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_2_task_2_ex.state)
        self.assertEqual(states.SUCCESS, wf_2_task_1_action_exs[0].state)
        self.assertEqual(states.SUCCESS, wf_2_task_2_action_exs[0].state)
        self.assertEqual(states.SUCCESS, wf_3_ex.state)
        self.assertEqual(2, len(wf_3_task_execs))
        self.assertEqual(states.SUCCESS, wf_3_task_1_ex.state)
        self.assertEqual(states.SUCCESS, wf_3_task_2_ex.state)
        self.assertEqual(states.SUCCESS, wf_3_task_1_action_exs[0].state)
        self.assertEqual(states.SUCCESS, wf_3_task_2_action_exs[0].state)



# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral.db.v2 import api as db_api
from mistral import exceptions as exc
from mistral.lang import parser as spec_parser
from mistral import utils
from mistral.workflow import states
from oslo_log import log as logging


STD_WF_PATH = 'resources/workflows'
LOG = logging.getLogger(__name__)


def register_standard_workflows(run_in_tx=True):
    LOG.debug("Registering standard workflows")
    workflow_paths = utils.get_file_list(STD_WF_PATH)

    for wf_path in workflow_paths:
        workflow_definition = open(wf_path).read()

        create_workflows(
            workflow_definition,
            scope='public',
            is_system=True,
            run_in_tx=run_in_tx,
            namespace=''
        )


def _clear_system_workflow_db():
    db_api.delete_workflow_definitions(is_system=True)


def sync_db():
    LOG.debug("Syncing db")
    with db_api.transaction():
        _clear_system_workflow_db()
        register_standard_workflows(run_in_tx=False)


def create_workflows(definition, scope='private', is_system=False,
                     run_in_tx=True, namespace=''):
    LOG.debug("Creating workflows")
    wf_list_spec = spec_parser.get_workflow_list_spec_from_yaml(
        definition, validate=True
    )
    db_wfs = []

    if run_in_tx:
        with db_api.transaction():
            _append_all_workflows(
                definition,
                is_system,
                scope,
                namespace,
                wf_list_spec,
                db_wfs
            )
    else:
        _append_all_workflows(
            definition,
            is_system,
            scope,
            namespace,
            wf_list_spec,
            db_wfs
        )

    return db_wfs


def _append_all_workflows(definition, is_system, scope, namespace,
                          wf_list_spec, db_wfs):
    for wf_spec in wf_list_spec.get_workflows():
        db_wfs.append(
            _create_workflow(
                wf_spec,
                definition,
                scope,
                namespace,
                is_system
            )
        )


def update_workflows(definition, scope='private', identifier=None,
                     namespace=''):
    LOG.debug("Updating workflows")

    wf_list_spec = spec_parser.get_workflow_list_spec_from_yaml(
        definition, validate=True
    )
    wfs = wf_list_spec.get_workflows()

    if identifier and len(wfs) > 1:
        raise exc.InputException(
            "More than one workflows are not supported for "
            "update with identifier. [identifier: %s]" %
            identifier
        )

    db_wfs = []

    with db_api.transaction():
        for wf_spec in wf_list_spec.get_workflows():
            db_wfs.append(_update_workflow(
                wf_spec,
                definition,
                scope,
                namespace=namespace,
                identifier=identifier
            ))

    return db_wfs


def update_workflow_execution_env(wf_ex, env):
    if not env:
        return wf_ex

    if wf_ex.state not in [states.IDLE, states.PAUSED, states.ERROR]:
        raise exc.NotAllowedException(
            'Updating env to workflow execution is only permitted if '
            'it is in IDLE, PAUSED, or ERROR state.'
        )

    wf_ex.params['env'] = utils.merge_dicts(wf_ex.params['env'], env)

    return wf_ex


def _get_workflow_values(wf_spec, definition, scope, namespace=None,
                         is_system=False):
    values = {
        'name': wf_spec.get_name(),
        'tags': wf_spec.get_tags(),
        'definition': definition,
        'spec': wf_spec.to_dict(),
        'scope': scope,
        'namespace': namespace,
        'is_system': is_system
    }

    return values


def _create_workflow(wf_spec, definition, scope, namespace, is_system):
    return db_api.create_workflow_definition(
        _get_workflow_values(wf_spec, definition, scope, namespace, is_system)
    )


def _update_workflow(wf_spec, definition, scope, identifier=None,
                     namespace=''):
    values = _get_workflow_values(wf_spec, definition, scope, namespace)

    return db_api.update_workflow_definition(
        identifier if identifier else values['name'],
        values
    )



# Copyright 2014 - Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_config import cfg

from mistral.db.v2 import api as db_api
from mistral.services import workflows as wf_service
from mistral.tests.unit.engine import base
from mistral.workflow import states

# Use the set_default method to set value otherwise in certain test cases
# the change in value is not permanent.
cfg.CONF.set_default('auth_enable', False, group='pecan')

WF = """
---
version: '2.0'

wf:
  type: direct

  input:
    - num1
    - num2

  output:
    result: <% $.result %>

  tasks:
    task1:
      action: std.echo output=<% $.num1 %>
      publish:
        result1: <% task(task1).result %>
      on-complete:
        - task3

    task2:
      action: std.echo output=<% $.num2 %>
      publish:
        result2: <% task(task2).result %>
      on-complete:
        - task3

    task3:
      description: |
        This task doesn't "action" or "workflow" property. It works as "no-op"
        task and serves just a decision point in the workflow.
      join: all
      on-complete:
        - task4: <% $.num1 + $.num2 = 2 %>
        - task5: <% $.num1 + $.num2 = 3 %>

    task4:
      action: std.echo output=4
      publish:
        result: <% task(task4).result %>

    task5:
      action: std.echo output=5
      publish:
        result: <% task(task5).result %>
"""


class NoopTaskEngineTest(base.EngineTestCase):
    def test_noop_task1(self):
        wf_service.create_workflows(WF)

        # Start workflow.
        wf_ex = self.engine.start_workflow(
            'wf',
            wf_input={'num1': 1, 'num2': 1}
        )

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            wf_output = wf_ex.output
            tasks = wf_ex.task_executions

        self.assertEqual(4, len(tasks))

        task1 = self._assert_single_item(tasks, name='task1')
        task2 = self._assert_single_item(tasks, name='task2')
        task3 = self._assert_single_item(tasks, name='task3')
        task4 = self._assert_single_item(tasks, name='task4')

        self.assertEqual(states.SUCCESS, task1.state)
        self.assertEqual(states.SUCCESS, task2.state)
        self.assertEqual(states.SUCCESS, task3.state)
        self.assertEqual(states.SUCCESS, task4.state)

        self.assertDictEqual({'result': 4}, wf_output)

    def test_noop_task2(self):
        wf_service.create_workflows(WF)

        # Start workflow.
        wf_ex = self.engine.start_workflow(
            'wf',
            wf_input={'num1': 1, 'num2': 2}
        )

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            wf_output = wf_ex.output
            tasks = wf_ex.task_executions

        self.assertEqual(4, len(tasks))

        task1 = self._assert_single_item(tasks, name='task1')
        task2 = self._assert_single_item(tasks, name='task2')
        task3 = self._assert_single_item(tasks, name='task3')
        task5 = self._assert_single_item(tasks, name='task5')

        self.assertEqual(states.SUCCESS, task1.state)
        self.assertEqual(states.SUCCESS, task2.state)
        self.assertEqual(states.SUCCESS, task3.state)
        self.assertEqual(states.SUCCESS, task5.state)

        self.assertDictEqual({'result': 5}, wf_output)



# Copyright (c) 2017 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools
import random
import six

import oslo_messaging as messaging


class KombuHosts(object):
    def __init__(self, conf):
        transport_url = messaging.TransportURL.parse(conf, conf.transport_url)

        if transport_url.hosts:
            self.virtual_host = transport_url.virtual_host
            self.hosts = transport_url.hosts
        else:
            self.virtual_host = conf.oslo_messaging_rabbit.rabbit_virtual_host
            self.hosts = []

            username = conf.oslo_messaging_rabbit.rabbit_userid
            password = conf.oslo_messaging_rabbit.rabbit_password

            for host in conf.oslo_messaging_rabbit.rabbit_hosts:
                hostname, port = host.split(':')

                self.hosts.append(messaging.TransportHost(
                    hostname,
                    int(port),
                    username,
                    password
                ))

        if len(self.hosts) > 1:
            random.shuffle(self.hosts)

        self._hosts_cycle = itertools.cycle(self.hosts)

    def get_host(self):
        return six.next(self._hosts_cycle)



# Copyright 2015 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral.lang import parser as spec_parser
from mistral.lang.v2 import workflows
from mistral.workflow import states


class WorkflowCommand(object):
    """Workflow command.

    A set of workflow commands form a communication protocol between workflow
    controller and its clients. When workflow controller makes a decision about
    how to continue a workflow it returns a set of commands so that a caller
    knows what to do next.
    """

    def __init__(self, wf_ex, wf_spec, task_spec, ctx, triggered_by=None):
        self.wf_ex = wf_ex
        self.wf_spec = wf_spec
        self.task_spec = task_spec
        self.ctx = ctx or {}
        self.triggered_by = triggered_by

    def to_dict(self):
        return {
            'task_name': self.task_spec.get_name(),
            'ctx': self.ctx,
            'triggered_by': self.triggered_by
        }


class Noop(WorkflowCommand):
    """No-operation command."""

    def __repr__(self):
        return "NOOP [workflow=%s]" % self.wf_ex.name

    def to_dict(self):
        d = super(Noop, self).to_dict()

        d['cmd_name'] = 'noop'

        return d


class RunTask(WorkflowCommand):
    """Instruction to run a workflow task."""

    def __init__(self, wf_ex, wf_spec, task_spec, ctx, triggered_by=None):
        super(RunTask, self).__init__(
            wf_ex,
            wf_spec,
            task_spec,
            ctx,
            triggered_by=triggered_by
        )

        self.wait = False
        self.unique_key = None

    def is_waiting(self):
        return self.wait

    def get_unique_key(self):
        return self.unique_key

    def __repr__(self):
        return (
            "Run task [workflow=%s, task=%s, waif_flag=%s, triggered_by=%s]" %
            (
                self.wf_ex.name,
                self.task_spec.get_name(),
                self.wait,
                self.triggered_by
            )
        )

    def to_dict(self):
        d = super(RunTask, self).to_dict()

        d['cmd_name'] = self.task_spec.get_name()
        d['wait'] = self.wait
        d['unique_key'] = self.unique_key

        return d


class RunExistingTask(WorkflowCommand):
    """Command to run an existing workflow task."""

    def __init__(self, wf_ex, wf_spec, task_ex, reset=True, triggered_by=None):
        super(RunExistingTask, self).__init__(
            wf_ex,
            wf_spec,
            spec_parser.get_task_spec(task_ex.spec),
            task_ex.in_context,
            triggered_by=triggered_by
        )

        self.task_ex = task_ex
        self.reset = reset
        self.unique_key = task_ex.unique_key

    def to_dict(self):
        d = super(RunExistingTask, self).to_dict()

        d['cmd_name'] = 'run_existing_task'
        d['task_ex_id'] = self.task_ex.id
        d['reset'] = self.reset
        d['unique_key'] = self.unique_key

        return d


class SetWorkflowState(WorkflowCommand):
    """Instruction to change a workflow state."""

    def __init__(self, wf_ex, wf_spec, task_spec, ctx, new_state, msg=None,
                 triggered_by=None):
        super(SetWorkflowState, self).__init__(
            wf_ex,
            wf_spec,
            task_spec,
            ctx,
            triggered_by=triggered_by
        )

        self.new_state = new_state
        self.msg = msg

    def to_dict(self):
        d = super(SetWorkflowState, self).to_dict()

        d['new_state'] = self.new_state
        d['msg'] = self.msg

        return d


class FailWorkflow(SetWorkflowState):
    """Instruction to fail a workflow."""

    def __init__(self, wf_ex, wf_spec, task_spec, ctx, msg=None,
                 triggered_by=None):
        super(FailWorkflow, self).__init__(
            wf_ex,
            wf_spec,
            task_spec,
            ctx,
            states.ERROR,
            msg=msg,
            triggered_by=triggered_by
        )

    def __repr__(self):
        return "Fail [workflow=%s]" % self.wf_ex.name

    def to_dict(self):
        d = super(FailWorkflow, self).to_dict()

        d['cmd_name'] = 'fail'

        return d


class SucceedWorkflow(SetWorkflowState):
    """Instruction to succeed a workflow."""

    def __init__(self, wf_ex, wf_spec, task_spec, ctx, msg=None,
                 triggered_by=None):
        super(SucceedWorkflow, self).__init__(
            wf_ex,
            wf_spec,
            task_spec,
            ctx,
            states.SUCCESS,
            msg=msg,
            triggered_by=triggered_by
        )

    def __repr__(self):
        return "Succeed [workflow=%s]" % self.wf_ex.name

    def to_dict(self):
        d = super(SucceedWorkflow, self).to_dict()

        d['cmd_name'] = 'succeed'

        return d


class PauseWorkflow(SetWorkflowState):
    """Instruction to pause a workflow."""

    def __init__(self, wf_ex, wf_spec, task_spec, ctx, msg=None,
                 triggered_by=None):
        super(PauseWorkflow, self).__init__(
            wf_ex,
            wf_spec,
            task_spec,
            ctx,
            states.PAUSED,
            msg=msg,
            triggered_by=triggered_by
        )

    def __repr__(self):
        return "Pause [workflow=%s]" % self.wf_ex.name

    def to_dict(self):
        d = super(PauseWorkflow, self).to_dict()

        d['cmd_name'] = 'pause'

        return d


ENGINE_CMD_CLS = {
    workflows.NOOP_COMMAND: Noop,
    workflows.FAIL_COMMAND: FailWorkflow,
    workflows.SUCCEED_COMMAND: SucceedWorkflow,
    workflows.PAUSE_COMMAND: PauseWorkflow
}


def get_command_class(cmd_name):
    return ENGINE_CMD_CLS[cmd_name] if cmd_name in ENGINE_CMD_CLS else None


# TODO(rakhmerov): IMO the way how we instantiate commands is weird.
# If we look at how we implement the logic of saving commands to
# dicts (to_dict) and restoring back from dicts then we'll see
# the lack of symmetry and unified way to do that depending on a
# command. Also RunExistingTask turns to be a special case that
# is not processed with this method at all. Might be a 'bad smell'.
# This all makes me think that we need to do some refactoring here.
def create_command(cmd_name, wf_ex, wf_spec, task_spec, ctx,
                   params=None, triggered_by=None):
    cmd_cls = get_command_class(cmd_name) or RunTask

    if issubclass(cmd_cls, SetWorkflowState):
        return cmd_cls(
            wf_ex,
            wf_spec,
            task_spec,
            ctx,
            msg=params.get('msg'),
            triggered_by=triggered_by
        )
    else:
        return cmd_cls(
            wf_ex,
            wf_spec,
            task_spec,
            ctx,
            triggered_by=triggered_by
        )


def restore_command_from_dict(wf_ex, cmd_dict):
    cmd_name = cmd_dict['cmd_name']

    wf_spec = spec_parser.get_workflow_spec_by_execution_id(wf_ex.id)
    task_spec = wf_spec.get_tasks()[cmd_dict['task_name']]
    ctx = cmd_dict['ctx']
    params = {'msg': cmd_dict.get('msg')} if 'msg' in cmd_dict else None
    triggered_by = cmd_dict.get('triggered_by')

    return create_command(
        cmd_name,
        wf_ex,
        wf_spec,
        task_spec,
        ctx,
        params,
        triggered_by
    )



# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import six

from mistral.lang import types
from mistral.lang.v2 import base
from mistral.lang.v2 import on_clause
from mistral.lang.v2 import policies
from mistral.lang.v2 import retry_policy


# TODO(rakhmerov): This specification should be broken into two separate
# specs for direct and reverse workflows. It's weird to combine them into
# one because they address different use cases.


class TaskDefaultsSpec(base.BaseSpec):
    # See http://json-schema.org
    _schema = {
        "type": "object",
        "properties": {
            "retry": retry_policy.RetrySpec.get_schema(),
            "wait-before": types.EXPRESSION_OR_POSITIVE_INTEGER,
            "wait-after": types.EXPRESSION_OR_POSITIVE_INTEGER,
            "timeout": types.EXPRESSION_OR_POSITIVE_INTEGER,
            "pause-before": types.EXPRESSION_OR_BOOLEAN,
            "concurrency": types.EXPRESSION_OR_POSITIVE_INTEGER,
            "on-complete": on_clause.OnClauseSpec.get_schema(),
            "on-success": on_clause.OnClauseSpec.get_schema(),
            "on-error": on_clause.OnClauseSpec.get_schema(),
            "safe-rerun": types.EXPRESSION_OR_BOOLEAN,
            "requires": {
                "oneOf": [types.NONEMPTY_STRING, types.UNIQUE_STRING_LIST]
            }
        },
        "additionalProperties": False
    }

    @classmethod
    def get_schema(cls, includes=['definitions']):
        return super(TaskDefaultsSpec, cls).get_schema(includes)

    def __init__(self, data, validate):
        super(TaskDefaultsSpec, self).__init__(data, validate)

        self._policies = self._group_spec(
            policies.PoliciesSpec,
            'retry',
            'wait-before',
            'wait-after',
            'timeout',
            'pause-before',
            'concurrency'
        )

        on_spec_cls = on_clause.OnClauseSpec

        self._on_complete = self._spec_property('on-complete', on_spec_cls)
        self._on_success = self._spec_property('on-success', on_spec_cls)
        self._on_error = self._spec_property('on-error', on_spec_cls)

        self._safe_rerun = data.get('safe-rerun')

        # TODO(rakhmerov): 'requires' should reside in a different spec for
        # reverse workflows.
        self._requires = data.get('requires', [])

    def validate_schema(self):
        super(TaskDefaultsSpec, self).validate_schema()

        self.validate_expr(self._data.get('safe-rerun', {}))

    def validate_semantics(self):
        # Validate YAQL expressions.
        self._validate_transitions(self._on_complete)
        self._validate_transitions(self._on_success)
        self._validate_transitions(self._on_error)

    def _validate_transitions(self, on_clause_spec):
        val = on_clause_spec.get_next() if on_clause_spec else []

        if not val:
            return

        [self.validate_expr(t)
            for t in ([val] if isinstance(val, six.string_types) else val)]

    def get_policies(self):
        return self._policies

    def get_on_complete(self):
        return self._on_complete

    def get_on_success(self):
        return self._on_success

    def get_on_error(self):
        return self._on_error

    def get_safe_rerun(self):
        return self._safe_rerun

    def get_requires(self):
        if isinstance(self._requires, six.string_types):
            return [self._requires]

        return self._requires



# Copyright 2013 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_config import cfg
from oslo_log import log as logging

from mistral import context as auth_ctx
from mistral.db.v2 import api as db_api
from mistral.db.v2.sqlalchemy import models
from mistral import exceptions as exc
from mistral import expressions as expr
from mistral.lang import parser as spec_parser
from mistral import utils
from mistral.utils import inspect_utils
from mistral.workflow import states

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class ContextView(dict):
    """Workflow context view.

    It's essentially an immutable composite structure providing fast lookup
    over multiple dictionaries w/o having to merge those dictionaries every
    time. The lookup algorithm simply iterates over provided dictionaries
    one by one and returns a value taken from the first dictionary where
    the provided key exists. This means that these dictionaries must be
    provided in the order of decreasing priorities.

    Note: Although this class extends built-in 'dict' it shouldn't be
    considered a normal dictionary because it may not implement all
    methods and account for all corner cases. It's only a read-only view.
    """

    def __init__(self, *dicts):
        super(ContextView, self).__init__()

        self.dicts = dicts or []

    def __getitem__(self, key):
        for d in self.dicts:
            if key in d:
                return d[key]

        raise KeyError(key)

    def get(self, key, default=None):
        for d in self.dicts:
            if key in d:
                return d[key]

        return default

    def __contains__(self, key):
        return any(key in d for d in self.dicts)

    def keys(self):
        keys = set()

        for d in self.dicts:
            keys.update(d.keys())

        return keys

    def items(self):
        return [(k, self[k]) for k in self.keys()]

    def values(self):
        return [self[k] for k in self.keys()]

    def iteritems(self):
        # NOTE: This is for compatibility with Python 2.7
        # YAQL converts output objects after they are evaluated
        # to basic types and it uses six.iteritems() internally
        # which calls d.items() in case of Python 2.7 and d.iteritems()
        # for Python 2.7
        return iter(self.items())

    def iterkeys(self):
        # NOTE: This is for compatibility with Python 2.7
        # See the comment for iteritems().
        return iter(self.keys())

    def itervalues(self):
        # NOTE: This is for compatibility with Python 2.7
        # See the comment for iteritems().
        return iter(self.values())

    def __len__(self):
        return len(self.keys())

    @staticmethod
    def _raise_immutable_error():
        raise exc.MistralError('Context view is immutable.')

    def __setitem__(self, key, value):
        self._raise_immutable_error()

    def update(self, E=None, **F):
        self._raise_immutable_error()

    def clear(self):
        self._raise_immutable_error()

    def pop(self, k, d=None):
        self._raise_immutable_error()

    def popitem(self):
        self._raise_immutable_error()

    def __delitem__(self, key):
        self._raise_immutable_error()


def evaluate_upstream_context(upstream_task_execs):
    published_vars = {}
    ctx = {}

    for t_ex in upstream_task_execs:
        # TODO(rakhmerov): These two merges look confusing. So it's a
        # temporary solution. There's still the bug
        # https://bugs.launchpad.net/mistral/+bug/1424461 that needs to be
        # fixed using context variable versioning.
        published_vars = utils.merge_dicts(
            published_vars,
            t_ex.published
        )

        utils.merge_dicts(ctx, evaluate_task_outbound_context(t_ex))

    return utils.merge_dicts(ctx, published_vars)


def _extract_execution_result(ex):
    if isinstance(ex, models.WorkflowExecution):
        return ex.output

    if ex.output:
        return ex.output['result']


def invalidate_task_execution_result(task_ex):
    for ex in task_ex.executions:
        ex.accepted = False


def get_task_execution_result(task_ex):
    execs = task_ex.executions
    execs.sort(
        key=lambda x: x.runtime_context.get('index')
    )

    results = [
        _extract_execution_result(ex)
        for ex in execs
        if hasattr(ex, 'output') and ex.accepted
    ]

    task_spec = spec_parser.get_task_spec(task_ex.spec)

    if task_spec.get_with_items():
        # TODO(rakhmerov): Smell: violation of 'with-items' encapsulation.
        with_items_ctx = task_ex.runtime_context.get('with_items')

        if with_items_ctx and with_items_ctx.get('count') > 0:
            return results
        else:
            return []

    return results[0] if len(results) == 1 else results


def publish_variables(task_ex, task_spec):
    if task_ex.state not in [states.SUCCESS, states.ERROR]:
        return

    wf_ex = task_ex.workflow_execution

    expr_ctx = ContextView(
        get_current_task_dict(task_ex),
        task_ex.in_context,
        get_workflow_environment_dict(wf_ex),
        wf_ex.context,
        wf_ex.input
    )

    if task_ex.name in expr_ctx:
        LOG.warning(
            'Shadowing context variable with task name while '
            'publishing: %s',
            task_ex.name
        )

    publish_spec = task_spec.get_publish(task_ex.state)

    if not publish_spec:
        return

    # Publish branch variables.
    branch_vars = publish_spec.get_branch()

    task_ex.published = expr.evaluate_recursively(branch_vars, expr_ctx)

    # Publish global variables.
    global_vars = publish_spec.get_global()

    utils.merge_dicts(
        task_ex.workflow_execution.context,
        expr.evaluate_recursively(global_vars, expr_ctx)
    )

    # TODO(rakhmerov):
    # 1. Publish atomic variables.
    # 2. Add the field "publish" in TaskExecution model similar to "published"
    #    but containing info as
    #    {'branch': {vars}, 'global': {vars}, 'atomic': {vars}}


def evaluate_task_outbound_context(task_ex):
    """Evaluates task outbound Data Flow context.

    This method assumes that complete task output (after publisher etc.)
    has already been evaluated.
    :param task_ex: DB task.
    :return: Outbound task Data Flow context.
    """

    # NOTE(rakhmerov): 'task_ex.in_context' has the SQLAlchemy specific
    # type MutableDict. So we need to create a shallow copy using dict(...)
    # initializer and use it. It's enough to be safe in order to manipulate
    # with entries of the result dictionary, like adding more entries.
    # However, we must not change values themselves because they are
    # shared between the original dictionary and the newly created.
    # It's better to avoid using the method copy.deepcopy() because on
    # dictionaries with many entries it significantly increases memory
    # footprint and reduces performance.
    in_context = (
        dict(task_ex.in_context)
        if task_ex.in_context is not None else {}
    )

    return utils.update_dict(in_context, task_ex.published)


def evaluate_workflow_output(wf_ex, wf_output, ctx):
    """Evaluates workflow output.

    :param wf_ex: Workflow execution.
    :param wf_output: Workflow output.
    :param ctx: Final Data Flow context (cause task's outbound context).
    """

    # Evaluate workflow 'output' clause using the final workflow context.
    ctx_view = ContextView(
        ctx,
        get_workflow_environment_dict(wf_ex),
        wf_ex.context,
        wf_ex.input
    )

    output = expr.evaluate_recursively(wf_output, ctx_view)

    # TODO(rakhmerov): Many don't like that we return the whole context
    # if 'output' is not explicitly defined.
    return output or ctx


def get_current_task_dict(task_ex):
    return {
        '__task_execution': {
            'id': task_ex.id,
            'name': task_ex.name
        }
    }


def add_openstack_data_to_context(wf_ex):
    wf_ex.context = wf_ex.context or {}

    if CONF.pecan.auth_enable:
        exec_ctx = auth_ctx.ctx()

        if exec_ctx:
            wf_ex.context.update({'openstack': exec_ctx.to_dict()})


def add_execution_to_context(wf_ex):
    wf_ex.context = wf_ex.context or {}

    wf_ex.context['__execution'] = {'id': wf_ex.id}


def add_workflow_variables_to_context(wf_ex, wf_spec):
    wf_ex.context = wf_ex.context or {}

    # The context for calculating workflow variables is workflow input
    # and other data already stored in workflow initial context.
    ctx_view = ContextView(
        get_workflow_environment_dict(wf_ex),
        wf_ex.context,
        wf_ex.input
    )

    wf_vars = expr.evaluate_recursively(wf_spec.get_vars(), ctx_view)

    utils.merge_dicts(wf_ex.context, wf_vars)


def evaluate_object_fields(obj, context):
    fields = inspect_utils.get_public_fields(obj)

    evaluated_fields = expr.evaluate_recursively(fields, context)

    for k, v in evaluated_fields.items():
        setattr(obj, k, v)


def get_workflow_environment_dict(wf_ex):
    if not wf_ex:
        return {}

    if wf_ex.root_execution_id:
        return get_workflow_environment_dict(
            db_api.get_workflow_execution(wf_ex.root_execution_id)
        )

    env_dict = wf_ex.params['env'] if 'env' in wf_ex.params else {}

    return {'__env': env_dict}



# Copyright 2016 NEC Corporation. All rights reserved.
# Copyright 2018 OVH SAS. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import datetime

import mock

from mistral.db.v2 import api as db_api
from mistral.db.v2.sqlalchemy import models
from mistral.tests.unit.api import base
from mistral.tests.unit.mstrlfixtures import policy_fixtures

MOCK_DELETE = mock.MagicMock(return_value=None)

WF_DEFINITION = """
---
version: '2.0'

flow:
  type: direct
  input:
    - param1

  tasks:
    task1:
      action: std.echo output="Hi"
"""
WF_DB = models.WorkflowDefinition(
    id='123e4567-e89b-12d3-a456-426655440000',
    name='flow',
    definition=WF_DEFINITION,
    created_at=datetime.datetime(1970, 1, 1),
    updated_at=datetime.datetime(1970, 1, 1),
    spec={'input': ['param1']}
)
MOCK_WF = mock.MagicMock(return_value=WF_DB)


class TestWorkflowPolicy(base.APITest):
    """Test workflow related policies

    Policies to test:
    - workflows:create
    - workflows:delete
    - workflows:get
    - workflows:list
    - workflows:list:all_projects
    - workflows:publicize (on POST & PUT)
    - workflows:update
    """

    def setUp(self):
        self.policy = self.useFixture(policy_fixtures.PolicyFixture())
        super(TestWorkflowPolicy, self).setUp()

    @mock.patch.object(db_api, "create_workflow_definition")
    def test_workflow_create_not_allowed(self, mock_obj):
        self.policy.change_policy_definition(
            {"workflows:create": "role:FAKE"}
        )
        resp = self.app.post(
            '/v2/workflows',
            WF_DEFINITION,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(403, resp.status_int)

    @mock.patch.object(db_api, "create_workflow_definition")
    def test_workflow_create_allowed(self, mock_obj):
        self.policy.change_policy_definition(
            {"workflows:create": "role:FAKE or rule:admin_or_owner"}
        )
        resp = self.app.post(
            '/v2/workflows',
            WF_DEFINITION,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(201, resp.status_int)

    @mock.patch.object(db_api, "create_workflow_definition")
    def test_workflow_create_public_not_allowed(self, mock_obj):
        self.policy.change_policy_definition({
            "workflows:create": "role:FAKE or rule:admin_or_owner",
            "workflows:publicize": "role:FAKE"
        })
        resp = self.app.post(
            '/v2/workflows?scope=public',
            WF_DEFINITION,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(403, resp.status_int)

    @mock.patch.object(db_api, "create_workflow_definition")
    def test_workflow_create_public_allowed(self, mock_obj):
        self.policy.change_policy_definition({
            "workflows:create": "role:FAKE or rule:admin_or_owner",
            "workflows:publicize": "role:FAKE or rule:admin_or_owner"
        })
        resp = self.app.post(
            '/v2/workflows?scope=public',
            WF_DEFINITION,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(201, resp.status_int)

    @mock.patch.object(db_api, "delete_workflow_definition", MOCK_DELETE)
    @mock.patch.object(db_api, "get_workflow_definition", MOCK_WF)
    def test_workflow_delete_not_allowed(self):
        self.policy.change_policy_definition(
            {"workflows:delete": "role:FAKE"}
        )
        resp = self.app.delete(
            '/v2/workflows/123',
            expect_errors=True
        )

        self.assertEqual(403, resp.status_int)

    @mock.patch.object(db_api, "delete_workflow_definition", MOCK_DELETE)
    @mock.patch.object(db_api, "get_workflow_definition", MOCK_WF)
    def test_workflow_delete_allowed(self):
        self.policy.change_policy_definition(
            {"workflows:delete": "role:FAKE or rule:admin_or_owner"}
        )
        resp = self.app.delete(
            '/v2/workflows/123',
            expect_errors=True
        )

        self.assertEqual(204, resp.status_int)

    @mock.patch.object(db_api, "get_workflow_definition", MOCK_WF)
    def test_workflow_get_not_allowed(self):
        self.policy.change_policy_definition(
            {"workflows:get": "role:FAKE"}
        )
        resp = self.app.get(
            '/v2/workflows/123',
            expect_errors=True
        )

        self.assertEqual(403, resp.status_int)

    @mock.patch.object(db_api, "get_workflow_definition", MOCK_WF)
    def test_workflow_get_allowed(self):
        self.policy.change_policy_definition(
            {"workflows:get": "role:FAKE or rule:admin_or_owner"}
        )
        resp = self.app.get(
            '/v2/workflows/123',
            expect_errors=True
        )

        self.assertEqual(200, resp.status_int)

    def test_workflow_list_not_allowed(self):
        self.policy.change_policy_definition(
            {"workflows:list": "role:FAKE"}
        )
        resp = self.app.get(
            '/v2/workflows',
            expect_errors=True
        )

        self.assertEqual(403, resp.status_int)

    def test_workflow_list_allowed(self):
        self.policy.change_policy_definition(
            {"workflows:list": "role:FAKE or rule:admin_or_owner"}
        )
        resp = self.app.get(
            '/v2/workflows',
            expect_errors=True
        )

        self.assertEqual(200, resp.status_int)

    def test_workflow_list_all_not_allowed(self):
        self.policy.change_policy_definition({
            "workflows:list": "role:FAKE or rule:admin_or_owner",
            "workflows:list:all_projects": "role:FAKE"
        })
        resp = self.app.get(
            '/v2/workflows?all_projects=1',
            expect_errors=True
        )

        self.assertEqual(403, resp.status_int)

    def test_workflow_list_all_allowed(self):
        self.policy.change_policy_definition({
            "workflows:list": "role:FAKE or rule:admin_or_owner",
            "workflows:list:all_projects": "role:FAKE or rule:admin_or_owner"
        })
        resp = self.app.get(
            '/v2/workflows?all_projects=1',
            expect_errors=True
        )

        self.assertEqual(200, resp.status_int)

    @mock.patch.object(db_api, "update_workflow_definition")
    def test_workflow_update_not_allowed(self, mock_obj):
        self.policy.change_policy_definition(
            {"workflows:update": "role:FAKE"}
        )
        resp = self.app.put(
            '/v2/workflows',
            WF_DEFINITION,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(403, resp.status_int)

    @mock.patch.object(db_api, "update_workflow_definition")
    def test_workflow_update_allowed(self, mock_obj):
        self.policy.change_policy_definition(
            {"workflows:update": "role:FAKE or rule:admin_or_owner"}
        )
        resp = self.app.put(
            '/v2/workflows',
            WF_DEFINITION,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(200, resp.status_int)

    @mock.patch.object(db_api, "update_workflow_definition")
    def test_workflow_update_public_not_allowed(self, mock_obj):
        self.policy.change_policy_definition({
            "workflows:update": "role:FAKE or rule:admin_or_owner",
            "workflows:publicize": "role:FAKE"
        })
        resp = self.app.put(
            '/v2/workflows?scope=public',
            WF_DEFINITION,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(403, resp.status_int)

    @mock.patch.object(db_api, "update_workflow_definition")
    def test_workflow_update_public_allowed(self, mock_obj):
        self.policy.change_policy_definition({
            "workflows:update": "role:FAKE or rule:admin_or_owner",
            "workflows:publicize": "role:FAKE or rule:admin_or_owner"
        })
        resp = self.app.put(
            '/v2/workflows?scope=public',
            WF_DEFINITION,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(200, resp.status_int)



# Copyright 2013 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import copy

import mock
import sqlalchemy as sa

from mistral.db.v2 import api as db_api
from mistral.db.v2.sqlalchemy import models
from mistral import exceptions as exc
from mistral.services import workbooks
from mistral.tests.unit.api import base

WORKBOOK_DEF = """
---
version: '2.0'
name: 'test'

workflows:
  flow:
    type: direct
    tasks:
      task1:
        action: step param="Hello !"

actions:
  step:
    input:
      - param
    base: std.echo output="<% $.param %>""
"""

UPDATED_WORKBOOK_DEF = """
---
version: '2.0'
name: 'book'

workflows:
  flow:
    type: direct
    tasks:
      task1:
        action: step arg="Hello !"

actions:
  step:
    input:
      - param
    base: std.echo output="<% $.param %>""
"""

WORKBOOK = {
    'id': '123',
    'name': 'test',
    'definition': WORKBOOK_DEF,
    'tags': ['deployment', 'demo'],
    'scope': 'public',
    'created_at': '1970-01-01 00:00:00',
    'updated_at': '1970-01-01 00:00:00'
}

WB_WITH_NAMESPACE = {
    'id': '123',
    'name': 'test',
    'namespace': 'xyz',
    'definition': WORKBOOK_DEF,
    'tags': ['deployment', 'demo'],
    'scope': 'public',
    'created_at': '1970-01-01 00:00:00',
    'updated_at': '1970-01-01 00:00:00'
}

ACTION = {
    'id': '123e4567-e89b-12d3-a456-426655440000',
    'name': 'step',
    'is_system': False,
    'description': 'My super cool action.',
    'tags': ['test', 'v2'],
    'definition': ''
}

WF = {
    'id': '123e4567-e89b-12d3-a456-426655440000',
    'name': 'flow',
    'definition': '',
    'created_at': '1970-01-01 00:00:00',
    'updated_at': '1970-01-01 00:00:00',
}

ACTION_DB = models.ActionDefinition()
ACTION_DB.update(ACTION)

WORKBOOK_DB = models.Workbook()
WORKBOOK_DB.update(WORKBOOK)

WB_DB_WITH_NAMESPACE = models.Workbook(**WB_WITH_NAMESPACE)

WF_DB = models.WorkflowDefinition()
WF_DB.update(WF)

WORKBOOK_DB_PROJECT_ID = WORKBOOK_DB.get_clone()
WORKBOOK_DB_PROJECT_ID.project_id = '<default-project>'

UPDATED_WORKBOOK_DB = copy.copy(WORKBOOK_DB)
UPDATED_WORKBOOK_DB['definition'] = UPDATED_WORKBOOK_DEF
UPDATED_WORKBOOK = copy.deepcopy(WORKBOOK)
UPDATED_WORKBOOK['definition'] = UPDATED_WORKBOOK_DEF

WB_DEF_INVALID_MODEL_EXCEPTION = """
---
version: '2.0'
name: 'book'

workflows:
  flow:
    type: direct
    tasks:
      task1:
        action: std.echo output="Hi"
        workflow: wf1
"""

WB_DEF_DSL_PARSE_EXCEPTION = """
---
%
"""

WB_DEF_YAQL_PARSE_EXCEPTION = """
---
version: '2.0'
name: 'book'

workflows:
  flow:
    type: direct
    tasks:
      task1:
        action: std.echo output=<% * %>
"""

MOCK_WORKBOOK = mock.MagicMock(return_value=WORKBOOK_DB)
MOCK_WB_WITH_NAMESPACE = mock.MagicMock(return_value=WB_DB_WITH_NAMESPACE)
MOCK_WORKBOOKS = mock.MagicMock(return_value=[WORKBOOK_DB])
MOCK_UPDATED_WORKBOOK = mock.MagicMock(return_value=UPDATED_WORKBOOK_DB)
MOCK_DELETE = mock.MagicMock(return_value=None)
MOCK_EMPTY = mock.MagicMock(return_value=[])
MOCK_NOT_FOUND = mock.MagicMock(side_effect=exc.DBEntityNotFoundError())
MOCK_DUPLICATE = mock.MagicMock(side_effect=exc.DBDuplicateEntryError())


class TestWorkbooksController(base.APITest):
    @mock.patch.object(db_api, "get_workbook", MOCK_WORKBOOK)
    def test_get(self):
        resp = self.app.get('/v2/workbooks/123')

        self.assertEqual(200, resp.status_int)
        self.assertDictEqual(WORKBOOK, resp.json)

    @mock.patch.object(db_api, "get_workbook", MOCK_WB_WITH_NAMESPACE)
    def test_get_with_namespace(self):
        resp = self.app.get('/v2/workbooks/123?namespace=xyz')

        self.assertEqual(200, resp.status_int)
        self.assertDictEqual(WB_WITH_NAMESPACE, resp.json)

    @mock.patch.object(db_api, 'get_workbook')
    def test_get_operational_error(self, mocked_get):
        mocked_get.side_effect = [
            # Emulating DB OperationalError
            sa.exc.OperationalError('Mock', 'mock', 'mock'),
            WORKBOOK_DB  # Successful run
        ]

        resp = self.app.get('/v2/workbooks/123')

        self.assertEqual(200, resp.status_int)
        self.assertDictEqual(WORKBOOK, resp.json)

    @mock.patch.object(db_api, "get_workbook", MOCK_NOT_FOUND)
    def test_get_not_found(self):
        resp = self.app.get('/v2/workbooks/123', expect_errors=True)

        self.assertEqual(404, resp.status_int)

    @mock.patch.object(db_api, "get_workbook",
                       return_value=WORKBOOK_DB_PROJECT_ID)
    def test_get_within_project_id(self, mock_get):
        resp = self.app.get('/v2/workbooks/123')

        self.assertEqual(200, resp.status_int)
        self.assertTrue('project_id' in resp.json)

    @mock.patch.object(workbooks, "update_workbook_v2", MOCK_UPDATED_WORKBOOK)
    def test_put(self):
        resp = self.app.put(
            '/v2/workbooks',
            UPDATED_WORKBOOK_DEF,
            headers={'Content-Type': 'text/plain'}
        )

        self.assertEqual(200, resp.status_int)
        self.assertEqual(UPDATED_WORKBOOK, resp.json)

    @mock.patch.object(workbooks, "update_workbook_v2", MOCK_NOT_FOUND)
    def test_put_not_found(self):
        resp = self.app.put_json(
            '/v2/workbooks',
            UPDATED_WORKBOOK_DEF,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(404, resp.status_int)

    def test_put_invalid(self):
        resp = self.app.put(
            '/v2/workbooks',
            WB_DEF_INVALID_MODEL_EXCEPTION,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)
        self.assertIn("Invalid DSL", resp.body.decode())

    @mock.patch.object(db_api, "update_workbook")
    @mock.patch.object(db_api, "create_or_update_workflow_definition")
    @mock.patch.object(db_api, "create_or_update_action_definition")
    def test_put_public(self, mock_action, mock_wf, mock_wb):
        mock_wb.return_value = UPDATED_WORKBOOK_DB
        mock_wf.return_value = WF_DB
        mock_action.return_value = ACTION_DB

        resp = self.app.put(
            '/v2/workbooks?scope=public',
            UPDATED_WORKBOOK_DEF,
            headers={'Content-Type': 'text/plain'}
        )

        self.assertEqual(200, resp.status_int)
        self.assertDictEqual(UPDATED_WORKBOOK, resp.json)

        self.assertEqual("public", mock_wb.call_args[0][1]['scope'])
        self.assertEqual("public", mock_wf.call_args[0][1]['scope'])
        self.assertEqual("public", mock_action.call_args[0][1]['scope'])

    def test_put_wrong_scope(self):
        resp = self.app.put(
            '/v2/workbooks?scope=unique',
            UPDATED_WORKBOOK_DEF,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)
        self.assertIn("Scope must be one of the following", resp.body.decode())

    @mock.patch.object(workbooks, "create_workbook_v2", MOCK_WORKBOOK)
    def test_post(self):
        resp = self.app.post(
            '/v2/workbooks',
            WORKBOOK_DEF,
            headers={'Content-Type': 'text/plain'}
        )

        self.assertEqual(201, resp.status_int)
        self.assertEqual(WORKBOOK, resp.json)

    @mock.patch.object(workbooks, "create_workbook_v2", MOCK_WB_WITH_NAMESPACE)
    def test_post_namespace(self):

        namespace = 'xyz'
        resp = self.app.post(
            '/v2/workbooks?namespace=%s' % namespace,
            WORKBOOK_DEF,
            headers={'Content-Type': 'text/plain'}
        )

        self.assertEqual(201, resp.status_int)
        self.assertEqual(WB_WITH_NAMESPACE, resp.json)

    @mock.patch.object(workbooks, "create_workbook_v2", MOCK_DUPLICATE)
    def test_post_dup(self):
        resp = self.app.post(
            '/v2/workbooks',
            WORKBOOK_DEF,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(409, resp.status_int)

    def test_post_invalid(self):
        resp = self.app.post(
            '/v2/workbooks',
            WB_DEF_INVALID_MODEL_EXCEPTION,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)
        self.assertIn("Invalid DSL", resp.body.decode())

    @mock.patch.object(db_api, "create_workbook")
    @mock.patch.object(db_api, "create_or_update_workflow_definition")
    @mock.patch.object(db_api, "create_or_update_action_definition")
    def test_post_public(self, mock_action, mock_wf, mock_wb):
        mock_wb.return_value = WORKBOOK_DB
        mock_wf.return_value = WF_DB
        mock_action.return_value = ACTION_DB

        resp = self.app.post(
            '/v2/workbooks?scope=public',
            WORKBOOK_DEF,
            headers={'Content-Type': 'text/plain'}
        )

        self.assertEqual(201, resp.status_int)
        self.assertEqual(WORKBOOK, resp.json)

        self.assertEqual("public", mock_wb.call_args[0][0]['scope'])
        self.assertEqual("public", mock_wf.call_args[0][1]['scope'])
        self.assertEqual("public", mock_action.call_args[0][1]['scope'])

    def test_post_wrong_scope(self):
        resp = self.app.post(
            '/v2/workbooks?scope=unique',
            WORKBOOK_DEF,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(400, resp.status_int)
        self.assertIn("Scope must be one of the following", resp.body.decode())

    @mock.patch.object(db_api, "delete_workbook", MOCK_DELETE)
    def test_delete(self):
        resp = self.app.delete('/v2/workbooks/123')

        self.assertEqual(204, resp.status_int)

    @mock.patch.object(db_api, "delete_workbook", MOCK_NOT_FOUND)
    def test_delete_not_found(self):
        resp = self.app.delete('/v2/workbooks/123', expect_errors=True)

        self.assertEqual(404, resp.status_int)

    @mock.patch.object(db_api, "get_workbooks", MOCK_WORKBOOKS)
    def test_get_all(self):
        resp = self.app.get('/v2/workbooks')

        self.assertEqual(200, resp.status_int)

        self.assertEqual(1, len(resp.json['workbooks']))
        self.assertDictEqual(WORKBOOK, resp.json['workbooks'][0])

    @mock.patch.object(db_api, 'get_workbooks')
    def test_get_all_operational_error(self, mocked_get_all):
        mocked_get_all.side_effect = [
            # Emulating DB OperationalError
            sa.exc.OperationalError('Mock', 'mock', 'mock'),
            [WORKBOOK_DB]  # Successful run
        ]

        resp = self.app.get('/v2/workbooks')

        self.assertEqual(200, resp.status_int)

        self.assertEqual(1, len(resp.json['workbooks']))
        self.assertDictEqual(WORKBOOK, resp.json['workbooks'][0])

    @mock.patch.object(db_api, "get_workbooks", MOCK_EMPTY)
    def test_get_all_empty(self):
        resp = self.app.get('/v2/workbooks')

        self.assertEqual(200, resp.status_int)

        self.assertEqual(0, len(resp.json['workbooks']))

    def test_validate(self):
        resp = self.app.post(
            '/v2/workbooks/validate',
            WORKBOOK_DEF,
            headers={'Content-Type': 'text/plain'}
        )

        self.assertEqual(200, resp.status_int)
        self.assertTrue(resp.json['valid'])

    def test_validate_invalid_model_exception(self):
        resp = self.app.post(
            '/v2/workbooks/validate',
            WB_DEF_INVALID_MODEL_EXCEPTION,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(200, resp.status_int)
        self.assertFalse(resp.json['valid'])
        self.assertIn("Invalid DSL", resp.json['error'])

    def test_validate_dsl_parse_exception(self):
        resp = self.app.post(
            '/v2/workbooks/validate',
            WB_DEF_DSL_PARSE_EXCEPTION,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(200, resp.status_int)
        self.assertFalse(resp.json['valid'])
        self.assertIn("Definition could not be parsed", resp.json['error'])

    def test_validate_yaql_parse_exception(self):
        resp = self.app.post(
            '/v2/workbooks/validate',
            WB_DEF_YAQL_PARSE_EXCEPTION,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(200, resp.status_int)
        self.assertFalse(resp.json['valid'])
        self.assertIn("unexpected '*' at position 1",
                      resp.json['error'])

    def test_validate_empty(self):
        resp = self.app.post(
            '/v2/workbooks/validate',
            '',
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(200, resp.status_int)
        self.assertFalse(resp.json['valid'])
        self.assertIn("Invalid DSL", resp.json['error'])



# Copyright 2015 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
The intention of the module is providing various DB related lookup functions
 for more convenient usage within the workflow engine.

Some of the functions may provide caching capabilities.

WARNING: Oftentimes, persistent objects returned by the methods in this
module won't be attached to the current DB SQLAlchemy session because
they are returned from the cache and therefore they need to be used
carefully without trying to do any lazy loading etc.
These objects are also not suitable for re-attaching them to a session
in order to update their persistent DB state.
Mostly, they are useful for doing any kind of fast lookups with in order
to make some decision based on their state.
"""

import threading

import cachetools
from oslo_config import cfg

from mistral.db.v2 import api as db_api
from mistral.workflow import states


CONF = cfg.CONF


def _create_lru_cache_for_workflow_execution(wf_ex_id):
    return cachetools.LRUCache(maxsize=500)

# This is a two-level caching structure.
# First level: [<workflow execution id> -> <task execution cache>]
# Second level (task execution cache): [<task_name> -> <task executions>]
# The first level (by workflow execution id) allows to invalidate
# needed cache entry when the workflow gets completed.
_TASK_EX_CACHE = cachetools.LRUCache(
    maxsize=100,
    missing=_create_lru_cache_for_workflow_execution
)

_ACTION_DEF_CACHE = cachetools.TTLCache(
    maxsize=1000,
    ttl=CONF.engine.action_definition_cache_time  # 60 seconds by default
)

_TASK_EX_CACHE_LOCK = threading.RLock()
_ACTION_DEF_CACHE_LOCK = threading.RLock()


def find_action_definition_by_name(action_name):
    """Find action definition name.

    :param action_name: Action name.
    :return: Action definition (possibly a cached value).
    """
    with _ACTION_DEF_CACHE_LOCK:
        action_definition = _ACTION_DEF_CACHE.get(action_name)

    if action_definition:
        return action_definition

    action_definition = db_api.load_action_definition(action_name)

    with _ACTION_DEF_CACHE_LOCK:
        _ACTION_DEF_CACHE[action_name] = action_definition

    return action_definition


def find_task_executions_by_name(wf_ex_id, task_name):
    """Finds task executions by workflow execution id and task name.

    :param wf_ex_id: Workflow execution id.
    :param task_name: Task name.
    :return: Task executions (possibly a cached value).
    """
    with _TASK_EX_CACHE_LOCK:
        t_execs = _TASK_EX_CACHE[wf_ex_id].get(task_name)

    if t_execs:
        return t_execs

    t_execs = db_api.get_task_executions(
        workflow_execution_id=wf_ex_id,
        name=task_name,
        sort_keys=[]  # disable sorting
    )

    # We can cache only finished tasks because they won't change.
    all_finished = (
        t_execs and
        all([states.is_completed(t_ex.state) for t_ex in t_execs])
    )

    if all_finished:
        with _TASK_EX_CACHE_LOCK:
            _TASK_EX_CACHE[wf_ex_id][task_name] = t_execs

    return t_execs


def find_task_executions_by_spec(wf_ex_id, task_spec):
    return find_task_executions_by_name(wf_ex_id, task_spec.get_name())


def find_task_executions_by_specs(wf_ex_id, task_specs):
    res = []

    for t_s in task_specs:
        res = res + find_task_executions_by_spec(wf_ex_id, t_s)

    return res


def find_task_executions_with_state(wf_ex_id, state):
    return db_api.get_task_executions(
        workflow_execution_id=wf_ex_id,
        state=state
    )


def find_successful_task_executions(wf_ex_id):
    return find_task_executions_with_state(wf_ex_id, states.SUCCESS)


def find_error_task_executions(wf_ex_id):
    return find_task_executions_with_state(wf_ex_id, states.ERROR)


def find_cancelled_task_executions(wf_ex_id):
    return find_task_executions_with_state(wf_ex_id, states.CANCELLED)


def find_completed_task_executions(wf_ex_id):
    return db_api.get_completed_task_executions(workflow_execution_id=wf_ex_id)


def find_completed_task_executions_as_batches(wf_ex_id):
    return db_api.get_completed_task_executions_as_batches(
        workflow_execution_id=wf_ex_id
    )


def get_task_execution_cache_size():
    return len(_TASK_EX_CACHE)


def get_action_definition_cache_size():
    return len(_ACTION_DEF_CACHE)


def invalidate_cached_task_executions(wf_ex_id):
    with _TASK_EX_CACHE_LOCK:
        if wf_ex_id in _TASK_EX_CACHE:
            del _TASK_EX_CACHE[wf_ex_id]


def clear_caches():
    with _TASK_EX_CACHE_LOCK:
        _TASK_EX_CACHE.clear()

    with _ACTION_DEF_CACHE_LOCK:
        _ACTION_DEF_CACHE.clear()
